import slab
import numpy
from matplotlib import pyplot as plt
plt.ioff()


def test_create_hrtf():
    hrtf1 = slab.HRTF.kemar()
    assert hrtf1.data[0].fir == 'IR'
    hrtf2 = slab.HRTF.kemar()
    for i in range(len(hrtf1.data)):
        numpy.testing.assert_equal(hrtf1[i].data, hrtf2[i].data)
    del hrtf2
    for i in range(100):
        idx = numpy.random.choice(range(hrtf1.n_sources))
        data = hrtf1[idx]  # make HRTF from instance of filter
        numpy.testing.assert_raises(ValueError, slab.HRTF, data)
        source = hrtf1.sources.vertical_polar[idx]
        listener = numpy.random.randn(3)
        hrtf = slab.HRTF(data=data, sources=source, listener=listener)
        numpy.testing.assert_equal(hrtf.listener, listener)
        numpy.testing.assert_equal(hrtf.sources.vertical_polar, source[numpy.newaxis, ...].astype('float16'))
        numpy.testing.assert_equal(hrtf[0].data.flatten(), data.data[:, 0])
        numpy.testing.assert_equal(hrtf[1].data.flatten(), data.data[:, 1])
        idx = numpy.random.choice(range(hrtf1.n_sources), 10, replace=False)
        data = [hrtf1[i].data for i in idx]  # make HRTF from array
        data = numpy.dstack(data)
        data = numpy.transpose(data, axes=(2, 0, 1))
        sources = hrtf1.sources.cartesian[idx]
        hrtf = slab.HRTF(data=data, sources=sources, samplerate=hrtf.samplerate, datatype='FIR')
        assert hrtf.n_sources == data.shape[0]
        assert hrtf[0].n_samples == data.shape[1]
        assert hrtf[0].n_filters == data.shape[2]


def test_plot_hrtf():
    hrtf = slab.HRTF.kemar()
    for ear in ["left", "right", "both"]:
        if ear == "both":
            _, ax = plt.subplots(2)
        else:
            _, ax = plt.subplots(1)
        for kind in ["waterfall", "image"]:
            cone = numpy.random.uniform(-180, 180)
            sources = hrtf.cone_sources(cone=cone)
            # sources = hrtf.cone_sources(cone=numpy.random.uniform(-180, 180))
            hrtf.plot_tf(sourceidx=sources, kind=kind, ear=ear, axis=ax, show=False)


def test_diffuse_field():
    hrtf = slab.HRTF.kemar()
    dfs = hrtf.diffuse_field_avg()
    assert dfs.fir == 'IR'
    assert dfs.n_frequencies == hrtf.data[0].n_taps
    equalized = hrtf.diffuse_field_equalization()
    for i in range(hrtf.n_sources):
        _, mag_raw = hrtf.data[i].tf(show=False)
        _, mag_eq = equalized.data[i].tf(show=False)
        assert numpy.abs(mag_eq.mean()) < numpy.abs(mag_raw.mean())


# TODO: working per software level, but need to check if the testing logic makes sense
def test_cone_sources():  # this is not working properly!
    hrtf = slab.HRTF.kemar()
    sound = slab.Binaural.whitenoise(samplerate=hrtf.samplerate)
    for _ in range(10):
        cone = numpy.random.uniform(-1000, 1000)
        idx = hrtf.cone_sources(cone)
        filtered = [hrtf.apply(i, sound) for i in idx]
        ilds = [f.ild() for f in filtered]
        assert numpy.abs(numpy.diff(ilds)).max() < 20


def test_elevation_sources():
    hrtf = slab.HRTF.kemar()
    elevations = numpy.unique(hrtf.sources.vertical_polar[:, 1])
    elevations = numpy.concatenate([elevations, [-35, 3, 21]])
    for e in elevations:
        sources = hrtf.elevation_sources(e)
        if e in numpy.unique(hrtf.sources.vertical_polar[:, 1]):
            assert all(numpy.logical_or(hrtf.sources.vertical_polar[sources][:, 0] <= 90,
                                        hrtf.sources.vertical_polar[sources][:, 0] >= 270))
            assert all(hrtf.sources.vertical_polar[sources][:, 1] == e)
        else:
            assert len(sources) == 0


def test_tf_from_sources():
    hrtf = slab.HRTF.kemar()
    for _ in range(10):
        n_sources = numpy.random.randint(1, 100)
        n_bins = numpy.random.randint(100, 500)
        sources = numpy.random.choice(range(len(hrtf.sources.vertical_polar)), n_sources)
        tfs = hrtf.tfs_from_sources(sources, n_bins=n_bins)
        assert tfs.shape[0] == n_bins
        assert tfs.shape[1] == n_sources


def test_vsi():
    hrtf = slab.HRTF.kemar()
    vsi = hrtf.vsi()
    numpy.testing.assert_almost_equal(vsi, 0.73, decimal=2)
    vsis = []
    for _ in range(10):
        sources = hrtf.cone_sources(cone=numpy.random.uniform(-180, 180))
        vsis.append(hrtf.vsi(sources=sources))
    assert all(numpy.logical_and(numpy.array(vsis) > 0.4, numpy.array(vsis) < 1.1))


def test_plot_sources():
    hrtf = slab.HRTF.kemar()
    for _ in range(10):
        idx = numpy.random.choice(range(len(hrtf.sources.vertical_polar)), numpy.random.randint(10))
        hrtf.plot_sources(idx=idx, show=False)


def test_interpolate():
    hrtf = slab.HRTF.kemar()
    for _ in range(10):
        idx = numpy.random.choice(range(len(hrtf.sources.vertical_polar)))
        azi, ele = hrtf.sources.vertical_polar[idx][0:2]
        method = numpy.random.choice(['nearest', 'bary'])
        filt = hrtf.interpolate(azimuth=azi, elevation=ele, method=method)
        _, spec_interp = filt.tf(show=False)
        _, spec_origin = hrtf[idx].tf(show=False)
        nearer_channel = 0 if azi-180 < 0 else 1
        # assert numpy.corrcoef(spec_interp[:, nearer_channel], spec_origin[:, nearer_channel]).min() > 0.99
        # TODO: another RNG-related failures from time to time
        # right now make the criterion less stringent
        assert numpy.corrcoef(spec_interp[:, nearer_channel], spec_origin[:, nearer_channel]).min() > 0.975


def test_convert_coordinates():
    hrtf = slab.HRTF.kemar()
    vertical_polar = hrtf.sources.vertical_polar.astype('float64')
    cartesian_from_vertical = slab.HRTF._vertical_polar_to_cartesian(vertical_polar)
    interaural_from_vertical = slab.HRTF._vertical_polar_to_interaural_polar(vertical_polar)
    cartesian_from_interaural = slab.HRTF._interaural_polar_to_cartesian(interaural_from_vertical)
    numpy.testing.assert_almost_equal(cartesian_from_interaural, cartesian_from_vertical, decimal=3)
    vertical_from_cartesian = slab.HRTF._cartesian_to_vertical_polar(cartesian_from_vertical)
    numpy.testing.assert_almost_equal(vertical_from_cartesian, vertical_polar, decimal=3)


def test_estimate_hrtf():
    fs = 44800
    duration = 1.0
    signal1 = slab.Sound.whitenoise(duration=duration, samplerate=fs)
    n_sources = numpy.random.randint(360)
    azimuths = numpy.random.choice(range(0, 360), size=n_sources, replace=False)
    elevations = numpy.random.choice(range(-180, 180), size=n_sources, replace=False)
    recordings = []
    sources = []
    for i in range(len(azimuths)):
        sources.append([azimuths[i], elevations[i], 1])
        recordings.append(slab.Binaural.whitenoise(duration = duration, samplerate=fs))
    hrtf1 = slab.HRTF.estimate_hrtf(recordings, signal1, sources)
    assert hrtf1[0].samplerate == recordings[0].samplerate
    assert hrtf1[0].n_samples == fs / 2 + 1
    assert hrtf1[0].n_filters == 2
    assert hrtf1.n_sources == n_sources
    assert hrtf1.n_elevations == n_sources
    signal2 = slab.Sound.whitenoise(duration=duration-0.1, samplerate=fs-800)
    hrtf2 = slab.HRTF.estimate_hrtf(recordings, signal2, sources)
    assert hrtf1.samplerate == hrtf2.samplerate
    assert hrtf1[0].n_samples == hrtf2[0].n_samples

def test_create_room():
    for i in range(10):
        size = numpy.random.randint(2, 51, 3)
        listener = size / 2
        source_cart = numpy.array([numpy.random.randint(1,i) for i in size])
        source_pol = slab.HRTF._get_coordinates(source_cart-listener, 'cartesian')[1][0]
        source_pol[0] -= 180
        max_echos = numpy.random.randint(1, 250)
        room = slab.Room(size=size, listener=listener, source=source_pol, order=1, max_echos=max_echos)
        numpy.testing.assert_almost_equal(source_pol[1:], room.image_locs[0,1:], decimal=0)
    room = slab.Room(size=[3,3,3], listener=[1,1,1])  # lists work?
    print(room)  # __str__ ok?

def test_reverb_time():
    room = slab.Room()
    t = room.reverb_time()
    t = room.reverb_time([5,5,5])  # list ok?
    assert isinstance(t, float)
    for _ in range(10):
        size = numpy.random.randint(1, 100, 3)
        t = room.reverb_time(size=size)

def test_hrir():
    for i in range(5):
        size = numpy.random.randint(2, 51, 3)
        listener = size / 2
        source_cart = numpy.array([numpy.random.randint(1,i) for i in size])
        source_pol = slab.HRTF._get_coordinates(source_cart-listener, 'cartesian')[1][0]
        source_pol[0] -= 180
        max_echos = numpy.random.randint(1, 50)
        room = slab.Room(size=size, listener=listener, source=source_pol, order=1, max_echos=max_echos)
        hrir = room.hrir()
    for i in range(5):
        tail = room.reverb(t_reverb=numpy.random.random()*3)  # up to T60 = 3s
        hrir = room.hrir(reverb=tail)  # try hrir with premade tail
    assert isinstance(hrir, slab.Filter)
    for _ in range(5):
        trim = numpy.random.randint(1, hrir.n_samples)
        hrir_tmp = room.hrir(trim=trim)  # try integer trim
        hrir_tmp = room.hrir(trim=numpy.random.random())  # try float trim
    sig = slab.Sound.vowel(duration=0.2, samplerate=hrir.samplerate)
    out = hrir.apply(sig)



import pathlib
import tempfile
import numpy
import scipy.signal
import slab
tmpdir = pathlib.Path(tempfile.gettempdir())


def test_sound_generation():
    # numpy.ndarray | str | pathlib.Path | list
    for _ in range(20):
        data = numpy.ones([10, 2])
        sound = slab.Sound(data, samplerate=10)  # generate sound from array
        sound1 = slab.Sound([data, data], samplerate=10)  # generate sound from list of arrays
        sound2 = slab.Sound([sound, sound])  # generate from list of sounds
        # sound1 and sound 2 should be the same:
        numpy.testing.assert_equal(sound1.data, sound2.data)
        numpy.testing.assert_equal(sound1.times, sound2.times)
        assert sound1.samplerate == sound2.samplerate
        assert sound1.duration == sound2.duration
        assert sound1.duration == sound2.duration
        # test if saving the file and initializing from string / path works. The reading and writing of data
        # is tested in more detail in test_read_write()
        sound = slab.Sound(numpy.random.randn(1000, 2), samplerate=numpy.random.randint(100, 1000))
        sound.write(tmpdir/"sound.wav", normalise=False)
        loaded1 = slab.Sound(tmpdir/"sound.wav")
        loaded2 = slab.Sound(str(tmpdir/"sound.wav"))
        numpy.testing.assert_equal(loaded1.data, loaded2.data)
        numpy.testing.assert_equal(loaded1.times, loaded2.times)
    # test if .name attribute of loaded sound object is the path string
    assert loaded1.name == f"{tmpdir / 'sound.wav'}"

def test_read_write():
    for _ in range(20):
        for normalize in [True, False]:
            sound = slab.Sound(numpy.random.randn(1000, 2), samplerate=numpy.random.randint(100, 1000))
            if normalize is False:
                sound.data = sound.data / sound.data.max()
            sound.write(tmpdir / "sound.wav", normalise=True)
            loaded = slab.Sound(tmpdir/"sound.wav")
            loaded.level = sound.level
            numpy.testing.assert_almost_equal(sound.data, loaded.data, decimal=3)


def test_tone():
    for freq in range(50, 20000, 500):
        sound = slab.Sound.tone(duration=numpy.random.randint(1000, 5000), frequency=freq, samplerate=44100)
        Z, freqs = sound.spectrum(show=False)
        assert numpy.abs(freqs[numpy.where(Z == Z.max())[0][0]] - freq) < 50
        assert sound.name == f'tone_{str(freq)}'
    for freq in range(500, 5000, 200):
        harmonic = slab.Sound.harmoniccomplex(duration=numpy.random.randint(1000, 5000), f0=freq, samplerate=44100)
        Z, freqs = harmonic.spectrum(show=False)
        peaks = scipy.signal.find_peaks(Z.flatten())[0]
        peak_freqs = freqs[peaks]
        peak_freqs = peak_freqs/freq
        numpy.testing.assert_almost_equal(peak_freqs, numpy.linspace(1, len(peaks), len(peaks)), decimal=0)
        assert harmonic.name == f'harmonic_{str(freq)}'


def test_powerlawnoise():
    for _ in range(20):
        centroids = []
        for alpha in numpy.linspace(.5, 1., 3):
            sound = slab.Sound.powerlawnoise(alpha=alpha, samplerate=44100)
            assert sound.name == f'powerlawnoise_{str(alpha)}'
            centroids.append(sound.spectral_feature("centroid"))
        assert all([centroids[i] > centroids[i+1] for i in range(len(centroids)-1)])


def test_crossfade():
    import itertools
    samplerate = 44100
    noise_durations = [0.1, 0.5, 1.0]
    vowel_durations = [0.1, 0.5, 1.0]
    overlaps = [0.0, 0.01]
    combinations = itertools.product(noise_durations, vowel_durations, overlaps)
    for noise_dur, vowel_dur, overlap in combinations:
        noise = slab.Sound.whitenoise(duration=noise_dur, samplerate=samplerate)
        vowel = slab.Sound.vowel(duration=vowel_dur, samplerate=samplerate)
        expected_n_samples = int(noise.n_samples + vowel.n_samples*2 - ((samplerate*overlap)*2))
        noise2vowel = slab.Sound.crossfade(vowel, noise, vowel, overlap=overlap)
        assert noise2vowel.n_samples == expected_n_samples
        if overlap == 0:  # crossfade with overlap 0 should be the same as sequence
            noise2vowel_seq = slab.Sound.sequence(vowel, noise, vowel)
            assert all(noise2vowel.data == noise2vowel_seq.data)
    noise_x_noise = slab.Sound.crossfade(noise, noise)
    assert noise_x_noise.name == f'{noise.name}_x_{noise.name}'


def test_frames():
    for _ in range(20):
        frame_dur = numpy.random.randint(10, 5000)
        sound_dur = numpy.abs(numpy.random.randn())+0.1
        sound = slab.Sound.whitenoise(duration=sound_dur)
        window_centers = sound.frametimes(duration=frame_dur)
        windows = sound.frames(duration=frame_dur)
        for window, center in zip(windows, window_centers):
            center1 = window[frame_dur][0]
            center2 = sound[numpy.where(sound.times == center)[0][0]][0]
            numpy.testing.assert_almost_equal(center1, center2, decimal=1)



import copy
import numpy
import math
import sys
from slab.sound import Sound
from slab.signal import Signal
from slab.filter import Filter
from slab.hrtf import HRTF

class Binaural(Sound):
    """
    Class for working with binaural sounds, including ITD and ILD manipulation. Binaural inherits all sound
    generation functions  from the Sound class, but returns binaural signals. Recasting an object of class sound or
    sound with 1 or 3+ channels calls Sound.copychannel to return a binaural sound with two channels identical
    to the first channel of the original sound.

    Arguments:
        data (slab.Signal | numpy.ndarray | list | str): see documentation of slab.Sound for details. the `data` must
            have either one or two channels. If it has one, that channel is duplicated
        samplerate (int): samplerate in Hz, must only be specified when creating an instance from an array.
        name (str): A string label for the Sound object. The inbuilt sound generating functions will automatically
            set .name to the name of the method used. Useful for logging during experiments.

    Attributes:
        .left: the first data channel, containing the sound for the left ear.
        .right: the second data channel, containing the sound for the right ear
        .data: the data-array of the Sound object which has the shape `n_samples` x `n_channels`.
        .n_channels: the number of channels in `data`. Must be 2 for a binaural sound.
        .n_samples: the number of samples in `data`. Equals `duration` * `samplerate`.
        .duration: the duration of the sound in seconds. Equals `n_samples` / `samplerate`.
        .name: string label of the sound.
   """
    # instance properties
    def _set_left(self, other):
        if hasattr(other, 'samplerate'):  # probably an slab object
            self.data[:, 0] = other.data[:, 0]
        else:
            self.data[:, 0] = numpy.array(other)

    def _set_right(self, other):
        if hasattr(other, 'samplerate'):  # probably an slab object
            self.data[:, 1] = other.data[:, 0]
        else:
            self.data[:, 1] = numpy.array(other)

    left = property(fget=lambda self: Sound(self.channel(0)), fset=_set_left,
                    doc='The left channel for a stereo sound.')
    right = property(fget=lambda self: Sound(self.channel(1)), fset=_set_right,
                     doc='The right channel for a stereo sound.')

    def __init__(self, data, samplerate=None, name='unnamed'):
        if isinstance(data, (Sound, Signal)):
            if hasattr(data, 'name'):
                self.name = data.name
            else:
                self.name = name
            if data.n_channels == 1:  # if there is only one channel, duplicate it.
                self.data = numpy.tile(data.data, 2)
            elif data.n_channels == 2:
                self.data = data.data
            else:
                raise ValueError("Data must have one or two channel!")
            self.samplerate = data.samplerate
        elif isinstance(data, (list, tuple)): # list of Sounds
            if isinstance(data[0], (Sound, Signal)):
                if data[0].n_samples != data[1].n_samples:
                    raise ValueError('Sounds must have same number of samples!')
                if data[0].samplerate != data[1].samplerate:
                    raise ValueError('Sounds must have same samplerate!')
                super().__init__([data[0].data[:, 0], data[1].data[:, 0]], data[0].samplerate, name=data[0].name)
            else: # list of samples
                super().__init__(data, samplerate, name=name)
        elif isinstance(data, str): # file name
            super().__init__(data, samplerate)
            if self.n_channels == 1:
                self.data = numpy.tile(self.data, 2)  # duplicate channel if monaural file
        else: # anything but Sound, list, or file name
            super().__init__(data, samplerate, name=name)
            if self.n_channels == 1:
                self.data = numpy.tile(self.data, 2)  # duplicate channel if monaural file
        if self.n_channels != 2: # bail if unable to enforce 2 channels
            ValueError('Binaural sounds must have two channels!')

    def itd(self, duration=None, max_lag=0.001):
        """
        Either estimate the interaural time difference of the sound or generate a new sound with the specified
        interaural time difference. The resolution for computing the ITD is 1/samplerate seconds. A negative
        ITD value means that the right channel is delayed, meaning the sound source is to the left.

        Arguments:
            duration (None| int | float): Given None, the instance's ITD is computed. Given another value, a new sound
                with the desired interaural time difference in samples (given an integer) or seconds (given a float)
                is generated.
            max_lag (float): Maximum possible value for ITD estimation. Defaults to 1 millisecond which is barely
                outside the physiologically plausible range for humans. Is ignored if `duration` is specified.
        Returns:
             (int | slab.Binaural): The interaural time difference in samples or a copy of the instance with the
                specified interaural time difference.
        Examples::

            sound = slab.Binaural.whitenoise()
            lateral = sound.itd(duration=0.0005)  # generate a sound with 0.5 ms ITD
            lateral.itd()  # estimate the ITD of the sound
        """
        if duration is None:
            return self._get_itd(max_lag)
        out = copy.deepcopy(self)
        out.name = f'{str(duration)}-itd_{self.name}'
        return out._apply_itd(duration)

    def _get_itd(self, max_lag):
        max_lag = Sound.in_samples(max_lag, self.samplerate)
        xcorr = numpy.correlate(self.data[:, 0], self.data[:, 1], 'full')
        lags = numpy.arange(-max_lag, max_lag + 1)
        xcorr = xcorr[self.n_samples - 1 - max_lag:self.n_samples + max_lag]
        idx = numpy.argmax(xcorr)
        return lags[idx]

    def _apply_itd(self, duration):
        if duration == 0:
            return self  # nothing needs to be shifted
        if duration < 0:  # negative itds by convention shift to the left (i.e. delay right channel)
            channel = 1  # right
        else:
            channel = 0  # left
        return self.delay(duration=abs(duration), channel=channel)

    def ild(self, dB=None):
        """
        Either estimate the interaural level difference of the sound or generate a new sound with the specified
        interaural level difference. Negative ILD value means that the left channel is louder than the right
        channel, meaning that the sound source is to the left. The level difference is achieved by adding half the ILD
        to one channel and subtracting half from the other channel, so that the mean intensity remains constant.

        Arguments:
            dB (None | int | float): If None, estimate the sound's ITD. Given a value, a new sound is generated with
                the desired interaural level difference in decibels.
        Returns:
            (float | slab.Binaural): The sound's interaural level difference, or a new sound with the specified ILD.
        Examples::

            sig = slab.Binaural.whitenoise()
            lateral_right = sig.ild(3) # attenuate left channel by 1.5 dB and amplify right channel by the same amount
            lateral_left = sig.ild(-3) # vice-versa
        """
        if dB is None:
            return self.right.level - self.left.level
        out = copy.deepcopy(self)  # so that we can return a new sound
        level = numpy.mean(self.level)
        out_levels = (level - dB/2, level + dB/2)
        out.level = out_levels
        out.name = f'{str(dB)}-ild_{self.name}'
        return out

    def itd_ramp(self, from_itd=-6e-4, to_itd=6e-4):
        """
        Generate a sound with a linearly increasing or decreasing interaural time difference. This is achieved by sinc
        interpolation of one channel with a dynamic delay. The resulting virtual sound source moves left or right.

        Arguments:
            from_itd (float): interaural time difference in seconds at the start of the sound.
                Negative numbers correspond to sources to the left of the listener.
            to_itd (float): interaural time difference in seconds at the end of the sound.
        Returns:
            (slab.Binaural): a copy of the instance wit the desired ITD ramp.
        Examples::

            sig = slab.Binaural.whitenoise()
            moving = sig.itd_ramp(from_itd=-0.001, to_itd=0.01)
            moving.play()
        """
        out = copy.deepcopy(self)
        # make the ITD ramps
        left_ramp = numpy.linspace(from_itd / 2, to_itd / 2, self.n_samples)
        right_ramp = numpy.linspace(-from_itd / 2, -to_itd / 2, self.n_samples)
        if self.n_samples >= 8192:
            filter_length = 1024
        elif self.n_samples >= 512:
            filter_length = self.n_samples // 16 * 2  # 1/8th of n_samples, always even
        else:
            raise ValueError('Signal too short! (min 512 samples)')
        out = out.delay(duration=left_ramp, channel=0, filter_length=filter_length)
        out = out.delay(duration=right_ramp, channel=1, filter_length=filter_length)
        out.name = f'ild-ramp_{self.name}'
        return out

    def ild_ramp(self, from_ild=-50, to_ild=50):
        """
        Generate a sound with a linearly increasing or decreasing interaural level difference. The resulting
        virtual sound source moves to the left or right.

        Arguments:
            from_ild (int | float): interaural level difference in decibels at the start of the sound.
                Negative numbers correspond to sources to the left of the listener.
            to_ild (int | float): interaural level difference in decibels at the end of the sound.
        Returns:
            (slab.Binaural): a copy of the instance with the desired ILD ramp. Any previously existing level difference
                is removed.
        Examples::

            sig = slab.Binaural.whitenoise()
            moving = sig.ild_ramp(from_ild=-10, to_ild=10)
            move.play()
        """
        out = self.ild(0)  # set ild to zero
        # make ramps
        left_ramp = numpy.linspace(-from_ild / 2, -to_ild / 2, self.n_samples)
        right_ramp = numpy.linspace(from_ild / 2, to_ild / 2, self.n_samples)
        left_ramp = 10**(left_ramp/20.)
        right_ramp = 10**(right_ramp/20.)
        # multiply channels with ramps
        out.data[:, 0] *= left_ramp
        out.data[:, 1] *= right_ramp
        out.name = f'ild-ramp_{self.name}'
        return out

    @staticmethod
    def azimuth_to_itd(azimuth, frequency=2000, head_radius=8.75):
        """
        Compute the ITD for a sound source at a given azimuth. For frequencies >= 2 kHz the Woodworth (1962)
        formula is used. For frequencies <= 500 Hz the low-frequency approximation mentioned in Aronson and Hartmann
        (2014) is used. For frequencies in between, we interpolate linearly between the two formulas.

        Arguments:
            azimuth (int | float): The azimuth angle of the sound source, negative numbers refer to sources to the left.
            frequency (int | float): Frequency in Hz for which the ITD is estimated.
                Use the default for for sounds with a broadband spectrum.
            head_radius (int | float): Radius of the head in centimeters. The bigger the head, the larger the ITD.
        Returns:
            (float): The interaural time difference for a sound source at a given azimuth.
        Examples::

            # compute the ITD for a sound source 90 degrees to the left for a large head
            itd = slab.Binaural.azimuth_to_itd(-90, head_radius=10)
        """
        head_radius = head_radius / 100
        azimuth_radians = numpy.radians(azimuth)
        speed_of_sound = 344  # m/s
        itd_2000 = (head_radius / speed_of_sound) * \
            (azimuth_radians + numpy.sin(azimuth_radians))  # Woodworth
        itd_500 = (3 * head_radius / speed_of_sound) * numpy.sin(azimuth_radians)
        itd = numpy.interp(frequency, [500, 2000], [itd_500, itd_2000],
                           left=itd_500, right=itd_2000)
        return itd

    @staticmethod
    def azimuth_to_ild(azimuth, frequency=2000, ils=None):
        """
        Get the interaural level difference corresponding to a sound source at a given azimuth and frequency.

        Arguments:
            azimuth (int | float): The azimuth angle of the sound source, negative numbers refer to sources to the left.
            frequency (int | float): Frequency in Hz for which the ITD is estimated.
                Use the default for for sounds with a broadband spectrum.
            ils (dict | None): interaural level spectrum from which the ILD is taken. If None,
            `make_interaural_level_spectrum()` is called. For repeated use, it is better to generate and keep the ils in
            a variable to avoid re-computing it.
        Returns:
            (float): The interaural level difference for a sound source at a given azimuth in decibels.
        Examples::

            ils = slab.Binaural.make_interaural_level_spectrum() # using default KEMAR HRTF
            ild = slab.Binaural.azimuth_to_ild(-90, ils=ils) # ILD equivalent to 90 deg leftward source for KEMAR
        """
        if ils is None:
            ils = Binaural.make_interaural_level_spectrum()
        # interpolate levels at azimuth:
        level_diffs = ils['level_diffs']
        levels = [numpy.interp(azimuth, ils['azimuths'], level_diffs[i, :]) for i in range(level_diffs.shape[0])]
        return numpy.interp(frequency, ils['frequencies'], levels)*-1   # interpolate level difference at frequency

    def at_azimuth(self, azimuth=0, ils=None):
        """
        Convenience function for adding ITD and ILD corresponding to the given `azimuth` to the sound source.
        Values are obtained from azimuth_to_itd and azimuth_to_ild.
        Frequency parameters for these functions are generated from the centroid frequency of the sound.

        Arguments:
            azimuth (int | float): The azimuth angle of the sound source, negative numbers refer to sources to the left.
            ils (dict | None): interaural level spectrum from which the ILD is taken. If None,
            `make_interaural_level_spectrum()` is called. For repeated use, it is better to generate and keep the ils in
            a variable to avoid re-computing it.
        Returns:
            (slab.Binaural): a sound with the appropriate ITD and ILD applied

        """
        centroid = numpy.array(self.spectral_feature(feature='centroid')).mean()
        itd = Binaural.azimuth_to_itd(azimuth, frequency=centroid)
        ild = Binaural.azimuth_to_ild(azimuth, frequency=centroid, ils=ils)
        out = self.itd(duration=itd)
        out.name = f'{azimuth}-azi_{self.name}'
        return out.ild(dB=ild)

    def externalize(self, hrtf=None):
        """
        Convolve the sound with a smoothed HRTF to evoke the impression of an external sound source without adding
        directional information, see Kulkarni & Colburn (1998) for why that works.

        Arguments:
            hrtf (None | slab.HRTF): The HRTF to use. If None use the one from the MIT KEMAR mannequin. The sound
                source at zero azimuth and elevation is used for convolution so it has to be present in the HRTF.
        Returns:
            (slab.Binaural): externalized copy of the instance.
        """
        if hrtf is None:
            hrtf = HRTF.kemar()  # load KEMAR as default
        # get HRTF for [0,0] direction:
        idx_frontal = numpy.where((hrtf.sources.vertical_polar[:, 1] == 0) &
                                  (hrtf.sources.vertical_polar[:, 0] == 0))[0][0]
        if not idx_frontal.size: # idx_frontal is empty
            raise ValueError('No frontal direction [0,0] found in HRTF.')
        _, h = hrtf.data[idx_frontal].tf(channels=0, n_bins=12, show=False)  # get low-res version of HRTF spectrum
        h[0] = 1  # avoids low-freq attenuation in KEMAR HRTF (unproblematic for other HRTFs)
        resampled_signal = copy.deepcopy(self)
        # if sound and HRTF has different samplerates, resample the sound, apply the HRTF, and resample back:
        resampled_signal = resampled_signal.resample(hrtf.data[0].samplerate)  # resample to hrtf rate
        filt = Filter(10**(h/20), fir='TF', samplerate=hrtf.data[0].samplerate)
        out = filt.apply(resampled_signal)
        out = out.resample(self.samplerate)
        out.name = f'externalized_{self.name}'
        return out

    @staticmethod
    def make_interaural_level_spectrum(hrtf=None):
        """
        Compute the frequency band specific interaural intensity differences for all sound source azimuth's in
        a head-related transfer function. For every azimuth in the hrtf, the respective transfer function is applied
        to a sound. This sound is then divided into frequency sub-bands. The interaural level spectrum is the level
        difference between right and left for each of these sub-bands for each azimuth.
        When individual HRTFs are not avilable, the level spectrum of the KEMAR mannequin may be used (default).
        Note that the computation may take a few minutes. Save the level spectrum to avoid re-computation, for instance
        with pickle or numpy.save (see documentation on readthedocs for examples).

        Arguments:
            hrtf (None | slab.HRTF): The head-related transfer function used to compute the level spectrum. If None,
                use the recordings from the KEMAR mannequin.
        Returns:
            (dict): A dictionary with keys `samplerate`, `frequencies` [n], `azimuths` [m], and `level_diffs` [n x m],
                where `frequencies` lists the centres of sub-bands for which the level difference was computed, and
                `azimuths` lists the sound source azimuth's in the hrft. `level_diffs` is a matrix of the interaural
                level difference for each sub-band and azimuth.
        Examples::

            ils = slab.Binaural.make_interaural_level_spectrum()  # get the ils from the KEMAR recordings
            ils['samplerate'] # the sampling rate
            ils['frequencies'] # the sub-band frequencies
            ils['azimuths']  # the sound source azimuth's for which the level difference was calculated
            ils['level_diffs'][5, :]  # the level difference for each azimuth in the 5th sub-band
        """
        if not hrtf:
            hrtf = HRTF.kemar()  # load KEMAR by default
        # get the filters for the frontal horizontal arc
        idx = numpy.where((hrtf.sources.vertical_polar[:, 1] == 0) & (
            (hrtf.sources.vertical_polar[:, 0] <= 90) | (hrtf.sources.vertical_polar[:, 0] >= 270)))[0]
        # at this point, we could just get the transfer function of each filter with hrtf.data[idx[i]].tf(),
        # but it may be better to get the spectral left/right differences with ERB-spaced frequency resolution:
        azi = hrtf.sources.vertical_polar[idx, 0]
        # 270<azi<360 -> azi-360 to get negative angles on the left
        azi[azi >= 270] = azi[azi >= 270]-360
        sort = numpy.argsort(azi)
        fbank = Filter.cos_filterbank(samplerate=hrtf.samplerate, pass_bands=True)
        freqs = fbank.filter_bank_center_freqs()
        noise = Sound.pinknoise(duration=5., samplerate=hrtf.samplerate)
        ils = dict()
        ils['samplerate'] = hrtf.samplerate
        ils['frequencies'] = freqs
        ils['azimuths'] = azi[sort]
        ils['level_diffs'] = numpy.zeros((len(freqs), len(idx)))
        for n, i in enumerate(idx[sort]):  # put the level differences in order of increasing angle
            noise_filt = Binaural(hrtf.data[i].apply(noise))
            noise_bank_left = fbank.apply(noise_filt.left)
            noise_bank_right = fbank.apply(noise_filt.right)
            ils['level_diffs'][:, n] = noise_bank_right.level - noise_bank_left.level
        return ils

    def interaural_level_spectrum(self, azimuth, ils=None):
        """
        Apply an interaural level spectrum, corresponding to a sound sources azimuth, to a
        binaural sound. The interaural level spectrum consists of frequency specific interaural level differences
        which are computed from a head related transfer function (see the `make_interaural_level_spectrum()` method).
        The binaural sound is divided into frequency sub-bands and the levels of each sub-band are set according to
        the respective level in the interaural level spectrum. Then, the sub-bands are summed up again into one
        binaural sound.

        Arguments:
            azimuth (int | float): azimuth for which the interaural level spectrum is calculated.
            ils (dict): interaural level spectrum to apply.  If None, `make_interaural_level_spectrum()` is called.
            For repeated use, it is better to generate and keep the ils in a variable to avoid re-computing it.
        Returns:
            (slab.Binaural): A binaural sound with the interaural level spectrum corresponding to the given azimuth.
        Examples::

            noise = slab.Binaural.pinknoise(kind='diotic')
            ils = slab.Binaural.make_interaural_level_spectrum() # using default KEMAR HRTF
            noise.interaural_level_spectrum(azimuth=-45, ils=ils).play()
        """
        if ils is None:
            ils = Binaural.make_interaural_level_spectrum()
        ils_samplerate = ils['samplerate']
        original_samplerate = self.samplerate
        azis = ils['azimuths']
        level_diffs = ils['level_diffs']
        levels = numpy.array([numpy.interp(azimuth, azis, level_diffs[i, :]) for i in range(level_diffs.shape[0])])
        # resample the signal to the rate of the HRTF from which the filter was computed:
        resampled = self.resample(samplerate=ils_samplerate)
        fbank = Filter.cos_filterbank(length=resampled.n_samples, samplerate=ils_samplerate, pass_bands=True)
        subbands_left = fbank.apply(resampled.left)
        subbands_right = fbank.apply(resampled.right)
        # change subband levels:
        subbands_left.level = subbands_left.level + levels / 2
        subbands_right.level = subbands_right.level - levels / 2
        out_left = Filter.collapse_subbands(subbands_left, filter_bank=fbank)
        out_right = Filter.collapse_subbands(subbands_right, filter_bank=fbank)
        out = Binaural([out_left, out_right])
        out.name = f'ils_{self.name}'
        return out.resample(samplerate=original_samplerate)

    def drr(self, winlength=0.0025):
        """
        Calculate the direct-to-reverberant-ratio, DRR for the impulse input. This is calculated by
        DRR = 10 * log10( X(T0-C:T0+C)^2 / X(T0+C+1:end)^2 ), where X is the approximated integral of the impulse,
        T0 is the time of the direct impulse, and C=2.5ms (Zahorik, P., 2002: 'Direct-to-reverberant energy ratio
        sensitivity', The Journal of the Acoustical Society of America, 112, 2110-2117)

        Arguments:
            winlength (int | float): specifies the length of the direct sound window. This window is used to calculate
            the energy of impulse sound, starting from the position of the peak amplitude of the impulse.

        Returns:
            Direct-to-reverberation value of the input impulse measured in dB

        Return type:
            (float)
        """
        # convert winlength parameter to samples
        winlength = Sound.in_samples(winlength, self.samplerate)
        # winlength should take minimum 2 sample values
        winlength = max(2, winlength)
        if winlength > len(self.data):
            raise ValueError("'winlength' should be shorter than input sound.")
        elif winlength > 0.01 * self.samplerate:
            raise ValueError("'winlength' is suggested to be under 10ms.")
        if self.samplerate < 5000:
            raise ValueError("Sampling frequency is too low, it should be at least 5000 Hz.")
        correction = round(0.5 * self.samplerate / 1000 / 1000) # safety margin of 0.5ms
        # calculate drr for left channel
        impulse = self.data[:, 0]
        impulse_sq = numpy.square(impulse)
        # find the location (index) of the maximal peak in the impulse signal
        peak_index = impulse_sq.argmax()
        # calculate direct and reverberant energy values by integrating under their RMS curve
        win_start_index = max(0, peak_index - correction)
        win_end_index = peak_index + winlength
        direct = impulse[win_start_index: win_end_index]
        direct_e = numpy.trapz(numpy.square(direct))
        reverb = impulse[win_end_index + 1:]
        reverb_e = numpy.trapz(numpy.square(reverb))
        # Calculate DRR as a value in dB
        if direct_e == 0:
            raise ValueError("Direct energy is 0. Please check that your input parameters are reasonable.")
        elif reverb_e == 0:
            raise ValueError("Reverb energy is 0. Are you sure this is an impulse?")
        elif reverb_e < sys.float_info.min:
            raise ValueError("Reverb energy is too low. Please check that your input parameters are reasonable.")
        ratio = direct_e / reverb_e
        drr = 10 * math.log10(ratio)
        # Return output
        return drr

    @staticmethod
    def whitenoise(kind='diotic', **kwargs):
        """
        Generate binaural white noise. `kind`='diotic' produces the same noise samples in both channels,
        `kind`='dichotic' produces uncorrelated noise. The rest is identical to `slab.Sound.whitenoise`.
        """
        if kind == 'dichotic':
            out = Binaural(Sound.whitenoise(n_channels=2, **kwargs))
            out.left = Sound.whitenoise(n_channels=1, **kwargs)
        elif kind == 'diotic':
            out = Binaural(Sound.whitenoise(n_channels=2, **kwargs))
            out.left = out.right
        else:
            raise ValueError("kind must be 'dichotic' or 'diotic'.")
        out.name = f'{kind}-{out.name}'
        return out

    @staticmethod
    def pinknoise(kind='diotic', **kwargs):
        """
        Generate binaural pink noise. `kind`='diotic' produces the same noise samples in both channels,
        `kind`='dichotic' produces uncorrelated noise. The rest is identical to `slab.Sound.pinknoise`.
        """
        out = Binaural.powerlawnoise(alpha=1, kind=kind, **kwargs)
        out.name = f'{kind}-pinknoise'
        return out

    @staticmethod
    def powerlawnoise(kind='diotic', **kwargs):
        """
        Generate binaural power law noise. `kind`='diotic' produces the same noise samples in both channels,
        `kind`='dichotic' produces uncorrelated noise. The rest is identical to `slab.Sound.powerlawnoise`.
        """
        if kind == 'dichotic':
            out = Binaural(Sound.powerlawnoise(n_channels=2, **kwargs))
            out.left = Sound.powerlawnoise(n_channels=1, **kwargs)
        elif kind == 'diotic':
            out = Binaural(Sound.powerlawnoise(n_channels=2, **kwargs))
            out.left = out.right
        else:
            raise ValueError("kind must be 'dichotic' or 'diotic'.")
        out.name = f'{kind}-{out.name}'
        return out

    @staticmethod
    def irn(kind='diotic', **kwargs):
        """
        Generate iterated ripple noise (IRN). `kind`='diotic' produces the same noise samples in both channels,
        `kind`='dichotic' produces uncorrelated noise. The rest is identical to `slab.Sound.irn`.
        """
        out = Binaural(Sound.irn(n_channels=2, **kwargs))
        if kind == 'diotic':
            out.left = out.right
        else:
            raise ValueError("kind must be 'dichotic' or 'diotic'.")
        out.name = f'{kind}-{out.name}'
        return out

    @staticmethod
    def tone(**kwargs):
        """ Identical to slab.Sound.tone, but with two channels. """
        return Binaural(Sound.tone(n_channels=2, **kwargs))

    @staticmethod
    def dynamic_tone(**kwargs):
        """ Identical to slab.Sound.dynamic_tone, but with two channels. """
        return Binaural(Sound.dynamic_tone(n_channels=2, **kwargs))

    @staticmethod
    def harmoniccomplex(**kwargs):
        """ Identical to slab.Sound.harmoniccomplex, but with two channels. """
        return Binaural(Sound.harmoniccomplex(n_channels=2, **kwargs))

    @staticmethod
    def click(**kwargs):
        """ Identical to slab.Sound.click, but with two channels. """
        return Binaural(Sound.click(n_channels=2, **kwargs))

    @staticmethod
    def clicktrain(**kwargs):
        """ Identical to slab.Sound.clicktrain, but with two channels. """
        return Binaural(Sound.clicktrain(**kwargs))

    @staticmethod
    def chirp(**kwargs):
        """ Identical to slab.Sound.chirp, but with two channels. """
        return Binaural(Sound.chirp(**kwargs))

    @staticmethod
    def silence(**kwargs):
        """ Identical to slab.Sound.silence, but with two channels. """
        return Binaural(Sound.silence(**kwargs))

    @staticmethod
    def vowel(**kwargs):
        """ Identical to slab.Sound.vowel, but with two channels. """
        return Binaural(Sound.vowel(**kwargs))

    @staticmethod
    def multitone_masker(**kwargs):
        """ Identical to slab.Sound.multitone_masker, but with two channels. """
        return Binaural(Sound.multitone_masker(**kwargs))

    @staticmethod
    def equally_masking_noise(**kwargs):
        """ Identical to slab.Sound.erb_noise, but with two channels. """
        return Binaural(Sound.equally_masking_noise(**kwargs))



import itertools
import pathlib
import tempfile
import slab
import numpy
tmpdir = pathlib.Path(tempfile.gettempdir())

def test_low_high_pass():
    for i in range(10):
        sound = slab.Sound.whitenoise(duration=2.0)
        for kind, fir in itertools.product(["lp", "hp"], ['FIR', 'IR', 'TF']):
            edge_freq = numpy.random.uniform(100, 2000)
            length = numpy.random.randint(500, 5000)
            filt = slab.Filter.band(frequency=edge_freq, length=length, kind=kind, fir=fir)
            filt_sound = filt.apply(sound)
            Z, freqs = filt_sound.spectrum(show=False)
            idx = numpy.abs(freqs-edge_freq).argmin()
            if kind == "hp":
                suppressed = Z[0:idx]
            else:
                suppressed = Z[idx:]
            assert suppressed.max() < -25


def test_band_pass_stop():
    sound = slab.Sound.whitenoise(duration=2.0)
    for kind, fir in itertools.product(["bp", "bs"], ['FIR', 'IR', 'TF']):
        lower_edge_freq = numpy.random.uniform(100, 1000)
        higher_edge_freq = lower_edge_freq + numpy.random.uniform(100, 1000)
        length = numpy.random.randint(500, 5000)
        filt = slab.Filter.band(frequency=(lower_edge_freq, higher_edge_freq), length=length, kind=kind, fir=fir)
        filt_sound = filt.apply(sound)
        Z, freqs = filt_sound.spectrum(show=False)
        low_idx = numpy.abs(freqs - lower_edge_freq).argmin()
        high_idx = numpy.abs(freqs - higher_edge_freq).argmin()
        if kind == "bp":
            suppressed = numpy.concatenate([Z[0:low_idx], Z[high_idx:]])
        else:
            suppressed = Z[low_idx:high_idx]
        assert suppressed.max() < -30


def test_custom_band():
    sound = slab.Sound.whitenoise(duration=2.0, samplerate=44100)
    freqs = numpy.array([100., 800., 2000., 4300., 8000., 14500., 18000.])
    gains = [
        [0., 1., 0., 1., 0., 1., 0.],
        [1., 0., 1, 0., 1., 0., 1.],
        [0., 1., 0., 0., 1., 1., 0.],
        [1., 0., 1., 0., 0., 1., 0.]
    ]
    for i in range(10):
        for fir, gain in itertools.product(['FIR', 'IR', 'TF'], gains):
            freqs += numpy.random.uniform(1, 10, 7)
            freqs.sort()
            length = numpy.random.randint(500, 5000)
            filt = slab.Filter.band(frequency=list(freqs), gain=gain, length=length, fir=fir, samplerate=sound.samplerate)
            w, h = filt.tf(show=False)
            suppressed_freqs = freqs[numpy.where(numpy.array(gain) == 0.0)]
            idx = [numpy.abs(w-freq).argmin() for freq in suppressed_freqs]
            assert max(h[idx]) < -20


def test_cos_filterbank():
    for _ in range(10):
        sound = slab.Sound.whitenoise(duration=1.0, samplerate=44100)
        length = numpy.random.randint(1000, 5000)
        low_cutoff = numpy.random.randint(0, 500)
        high_cutoff = numpy.random.choice([numpy.random.randint(5000, 15000), None])
        n_filters = []
        for bandwidth in numpy.linspace(0.1, 0.9, 9):
            fbank = slab.Filter.cos_filterbank(length=length, bandwidth=bandwidth, low_cutoff=low_cutoff, high_cutoff=high_cutoff, pass_bands=False, samplerate=sound.samplerate)
            n_filters.append(fbank.n_filters)
            filtsound = fbank.apply(sound)
            assert filtsound.n_channels == fbank.n_filters
            assert filtsound.n_samples == sound.n_samples
        assert all([n_filters[i] >= n_filters[i+1] for i in range(len(n_filters)-1)])
        bandwidth = numpy.random.uniform(0.1, 0.9)
        fbank = slab.Filter.cos_filterbank(
                length=sound.n_samples, bandwidth=bandwidth, low_cutoff=low_cutoff,
                high_cutoff=high_cutoff, pass_bands=True, samplerate=sound.samplerate)
        filtsound = fbank.apply(sound)
        collapsed = slab.Filter.collapse_subbands(filtsound, fbank)
        numpy.testing.assert_almost_equal(sound.data, collapsed.data, decimal=-1)


def test_center_freqs():
    for _ in range(100):
        low_cutoff = numpy.random.randint(0, 500)
        high_cutoff = numpy.random.choice([numpy.random.randint(5000, 20000)])
        bandwidth = numpy.random.uniform(0.1, 0.7)
        center_freqs1, bandwidth1, erb_spacing1 = slab.Filter._center_freqs(low_cutoff, high_cutoff, bandwidth=bandwidth, pass_bands=False)
        center_freqs2, bandwidth2, erb_spacing2 = slab.Filter._center_freqs(low_cutoff, high_cutoff, bandwidth=bandwidth, pass_bands=True)
        assert all(center_freqs1 == center_freqs2[1:-1])
        assert len(center_freqs1) == len(center_freqs2)-2
        assert bandwidth1 == bandwidth2
        assert erb_spacing1 == erb_spacing2
        n_filters = len(center_freqs1)
        center_freqs3, bandwidth3, erb_spacing3 = slab.Filter._center_freqs(low_cutoff, high_cutoff, n_filters=n_filters, pass_bands=False)
        assert all(center_freqs1 == center_freqs3)
        assert bandwidth1 == bandwidth3
        assert erb_spacing1 == erb_spacing3


def test_equalization():
    for i in range(10):
        length = numpy.random.randint(1000, 5000)
        low_cutoff = numpy.random.randint(20, 2000)
        high_cutoff = numpy.random.randint(10000, 20000)
        sound = slab.Sound.pinknoise(samplerate=44100, duration=length/1000)
        # bandpass to reduce the difference at low and high frequencies
        sound = sound.filter([20, 18000], 'bp')
        freqs = numpy.logspace(numpy.log2(50), numpy.log2(18000), 10, base=2)
        gains = 2 ** numpy.random.uniform(-2, 2, freqs.shape)
        filt = slab.Filter.band(frequency=freqs.tolist(),
                                gain=gains.tolist(), samplerate=sound.samplerate)
        filtered = filt.apply(sound)
        fbank = slab.Filter.equalizing_filterbank(sound, filtered, low_cutoff=low_cutoff, high_cutoff=high_cutoff)
        equalized = fbank.apply(filtered)
        Z_equalized, _ = equalized.spectrum(low_cutoff, high_cutoff, show=False)
        Z_sound, _ = sound.spectrum(low_cutoff, high_cutoff, show=False)
        Z_filtered, _ = filtered.spectrum(low_cutoff, high_cutoff, show=False)
        # The spectral difference between original and equalized should be minimum
        # TODO: now only check the average. might make sense to also check the maximum
        # TODO: the edge effect is annoying. should think of a way to deal with it
        assert numpy.abs(Z_sound-Z_equalized)[100:-1000].mean() < 1


def test_load_save():
    for kind, freq in zip(["lp", "hp", "bs", "bp"], [
        numpy.random.uniform(100, 2000),
        numpy.random.uniform(100, 2000),
        (0+numpy.random.uniform(100, 2000), 2000+numpy.random.uniform(100, 2000)),
        (0 + numpy.random.uniform(100, 2000), 2000 + numpy.random.uniform(100, 2000))
    ]):
        for fir in ('FIR', 'IR', 'TF'):
            filt = slab.Filter.band(kind=kind, frequency=freq, fir=fir)
            filt.save(tmpdir/"filt.npy")
            loaded = slab.Filter.load(tmpdir/"filt.npy")
            numpy.testing.assert_equal(filt.data, loaded.data)
            numpy.testing.assert_equal(filt.times, loaded.times)
            assert filt.fir == loaded.fir
            assert filt.n_frequencies == loaded.n_frequencies
            assert filt.n_taps == loaded.n_taps



'''
Michaela's experiment:
Interference between room and voice processing
Use like this:
>>> from slab.experiments import room_voice_interference
>>> room_voice_interference.main_experiment('subject01')
Please see docstring of main_experiment() for details.
'''

import time
import pathlib
import collections
import numpy
import slab

# configuration
slab.Signal.set_default_samplerate(44100)
_results_file = None
slab.ResultsFile.results_folder = 'Results'
slab.psychoacoustics.input_method = 'keyboard' # or 'buttonbox'
# possible parameters:
rooms = tuple(range(40, 161, 8)) # (40, 48, 56, 64, 72, 80, 88, 96, 104, 112, 120, 128, 136, 144, 152, 160)
voices = (0.98, 1.029, 1.078, 1.127, 1.176, 1.225, 1.274, 1.323, 1.372, 1.421, 1.47)
itds = tuple(range(0, 401, 40)) # (0, 40, 80, 120, 160, 200, 240, 280, 320, 360, 400)
default_room = rooms[0] # 40
default_voice = voices[0] # 0.98
default_itd = itds[0] # 0
ISI_stairs = 0.15
_after_stim_pause = 0.3
word_list = ['Aertel', 'Apor', 'Aucke', 'Bohke', 'Dercke', 'Eika', 'Eukof', 'Felcke', 'Geke', 'Gelkat', 'Kansup', 'Kelpeit', 'Kirpe', 'Kitlu', 'Klamsup', 'Kontus', 'Lanrapf', 'Manzeb', 'Nukal', 'Pekus', 'Perkus', 'Raupfan', 'Reiwat', 'Repad', 'Retel', 'Schaujak', 'Seckuck', 'Sekaak', 'Stiecke', 'Subter', 'Trepfel', 'Tunsat', 'Verzung', 'Waatep', 'Wieken', 'Zeten']
stim_folder = pathlib.Path('..') / pathlib.Path('Stimuli') # set to correct path
condition = collections.namedtuple('condition', ['voice', 'room', 'itd', 'label']) # used to set parameters in interference_block

def jnd(condition, practise=False):
    '''
    Presents a staricase for a 2AFC task and returns the threshold.
    This threshold is used in the main experiment as jnd.
    condition ... 'room', voice', or 'itd'
    '''
    print('Three sounds are presented in each trial.')
    print('They are always different, but sometimes')
    if condition == 'room':
        print('one sound is played in a larger room,')
        print('and sometimes all three are played in the same room.')
        print('Was the larger room presented first, second, or third?')
    elif condition == 'voice':
        print('one is spoken by a different (larger) person,')
        print('and sometimes all three are spoken by the same person.')
        print('Was the larger person presented first, second, or third?')
    elif condition == 'itd':
        print('one is played from a different direction (slightly to the left),')
        print('and sometimes all three are played from straight ahead.')
        print('Was the sound slightly from the left played first, second, or third?')
    else:
        raise ValueError(f'Invalid condition {condition}.')
    print('Press 1 for first, 2 for second, 3 for third.')
    print('The difference will get more and more difficult to hear.')
    input('Press enter to start JND estimation...')
    repeat = 'r'
    condition_values = globals()[condition+'s'] # get the parameter list (vars rooms, voices, or itds) from condition string
    while repeat == 'r':
        # make a random, non-repeating list of words to present during the staircase
        word_seq = slab.Trialsequence(conditions=word_list, kind='infinite', label='word_seq')
        # define the staircase
        if practise:
            stairs = slab.Staircase(start_val=len(condition_values)-1, n_reversals=3,
                                step_sizes=[4, 3, 2], min_val=0, max_val=len(condition_values)-1, n_up=1, n_down=1, n_pretrials=0)
        else:
            stairs = slab.Staircase(start_val=len(condition_values)-4, n_reversals=15,
                                step_sizes=[4, 2], min_val=0, max_val=len(condition_values)-1, n_up=1, n_down=2, step_up_factor=1.5, n_pretrials=1) # should give approx. 70% hitrate
            _results_file.write(f'{condition} jnd:', tag='time')
        for trial in stairs:
            current = condition_values[int(trial)]
            # load stimuli
            word = next(word_seq)
            word2 = next(word_seq)
            word3 = next(word_seq)
            if condition == 'room':
                jnd_stim = slab.Sound(stim_folder / word  / f'{word}_SER{default_voice:.4g}_GPR168_{current}_{default_itd}.wav')
            elif condition == 'voice':
                jnd_stim = slab.Sound(stim_folder / word  / f'{word}_SER{current:.4g}_GPR168_{default_room}_{default_itd}.wav')
            elif condition == 'itd':
                jnd_stim = slab.Sound(stim_folder / word  / f'{word}_SER{default_voice:.4g}_GPR168_{default_room}_{current}.wav')
            default_stim1 = slab.Sound(stim_folder / word2 / f'{word2}_SER{default_voice:.4g}_GPR168_{default_room}_{default_itd}.wav')
            default_stim2 = slab.Sound(stim_folder / word3 / f'{word3}_SER{default_voice:.4g}_GPR168_{default_room}_{default_itd}.wav')
            stairs.present_afc_trial(jnd_stim, [default_stim1, default_stim2], isi=ISI_stairs, print_info=practise)
            if practise:
                stairs.plot()
        thresh = stairs.threshold()
        thresh_condition_value = condition_values[numpy.ceil(thresh).astype('int')]
        if practise:
            stairs.close_plot()
        else:
            print(f'room jnd: {round(thresh, ndigits=1)}')
            _results_file.write(repr(stairs), tag=f'stairs {condition}')
            _results_file.write(thresh, tag=f'jnd {condition}')
            _results_file.write(thresh_condition_value, tag=f'jnd condition value {condition}')
        repeat = input('Press enter to continue, "r" to repeat this threshold measurement.\n\n')
    return thresh_condition_value

# same-diff task, method of constant stimuli, 5 conditions, 40 reps (20 diff, 20 same)
def interference_block(jnd_room, jnd_voice, jnd_itd):
    '''
    Presents one condition block of the the interference test.
    Condition ... 'room', 'room+voice', 'room+itd', 'voice', or 'itd'
    default_room etc. ... the reference room, SER and ITD values.
    jnd_room etc. ... the room, SER and ITD values that are perceived as different from the default
                      (default value + measured jnd rounded to the nearest available stimulus)
    '''
    print('Three sounds are presented in each trial.')
    print('They are always different, but sometimes')
    print('one sound is played in a larger room,')
    print('and sometimes all three are played in the same room.')
    print('Was the larger room presented first, second, or third?')
    print('Press 1 for first, 2 for second, and 3 for third.')
    input('Press enter to start the test...')
    # set parameter values of conditions in named tuples -> list of these is used for slab.Trialsequence
    default = condition(voice=default_voice, room=default_room, itd=default_itd, label='default')
    room = condition(voice=default_voice, room=jnd_room, itd=default_itd, label='room')
    room_voice = condition(voice=jnd_voice, room=jnd_room, itd=default_itd, label='room_voice')
    room_itd = condition(voice=default_voice, room=jnd_room, itd=jnd_itd, label='room_itd')
    voice = condition(voice=jnd_voice, room=default_room, itd=default_itd, label='voice')
    itd = condition(voice=default_voice, room=default_room, itd=jnd_itd, label='itd')
    conditions = [default, room, room_voice, room_itd, voice, itd]
    trials = slab.Trialsequence(conditions=conditions, n_reps=10, kind='random_permutation')
    word_seq = slab.Trialsequence(conditions=word_list, kind='infinite', label='word_seq')
    hits = 0
    false_alarms = 0
    _results_file.write(f'interference block:', tag='time')
    for trial_parameters in trials:
        # load stimuli
        word  = next(word_seq)
        word2 = next(word_seq)
        word3 = next(word_seq)
        jnd_stim = slab.Sound(str(stim_folder / word  / word) + '_SER%.4g_GPR168_%i_%i.wav' % trial_parameters[:-1])
        default_stim1 = slab.Sound(str(stim_folder / word2 / word2) + '_SER%.4g_GPR168_%i_%i.wav' % default[:-1])
        default_stim2 = slab.Sound(str(stim_folder / word3 / word3) + '_SER%.4g_GPR168_%i_%i.wav' % default[:-1])
        trials.present_afc_trial(jnd_stim, [default_stim1, default_stim2], isi=ISI_stairs)
        response = trials.data[-1] # read out the last response
        if trial_parameters.label[:4] == 'room' and response: # hit!
            hits += 1
        elif trial_parameters.label[:3] in ['voi', 'itd'] and response: # false alarm!
            false_alarms += 1
        time.sleep(_after_stim_pause)
    hitrate = hits/trials.n_trials
    print(f'hitrate: {hitrate}')
    farate = false_alarms/trials.n_trials
    print(f'false alarm rate: {farate}')
    _results_file.write(repr(trials), tag='trials')

def main_experiment(subject=None, do_jnd=True, do_interference=True):
    '''
    Interference between room and voice processing.
    Pre-recorded voice recordings are presented in different simulated rooms (the large stimulus
    set is not included). Just-noticeable differences for changes in room volume and voice
    parameters (glottal pulse rate and vocal tract length) are first measured, then 3-alternative-
    forced-choice trial are presented with the reference in a larger room. Does a simultaneous voice
    change impede the detection of the room change?
    The experiment requires a set of recorded spoken word stimuli, each of which was offline manipulated
    to change speaker identity (using the STRAIGHT algorithm) and then run through a room acoustics
    simulation to add reverberation consistent with rooms of different sizes. The filenames of the recordings
    contain the word and the voice and room parameters, so that the correct file is loaded for presentation.

    This experiment showcases participant data handling, AFC trials, prerecorded stimulus handling, among others.
    '''
    global _results_file
    # set up the results file
    if not subject:
        subject = input('Enter subject code: ')
    _results_file = slab.ResultsFile(subject=subject)
    if do_jnd:
        jnd('room', practise=True)  # run the stairs practice for the room condition
        jnd_room = jnd('room') # mesure
        jnd('voice', practise=True)  # run the stairs practice for the room condition
        jnd_voice = jnd('voice')
        jnd('itd', practise=True)  # run the stairs practice for the room condition
        jnd_itd = jnd('itd')
    else: # need to get the jnds from the resultsfile
        prev = slab.ResultsFile.previous_file(subject=subject)
        jnd_room = slab.ResultsFile.read_file(prev, tag='jnd_room')
        if not jnd_room:
            raise ValueError('jnd_room not found in previous results file. Please run the full experiment.')
        jnd_voice = slab.ResultsFile.read_file(prev, tag='jnd_voice')
        if not jnd_voice:
            raise ValueError('jnd_voice not found in previous results file. Please run the full experiment.')
        jnd_itd = slab.ResultsFile.read_file(prev, tag='jnd_itd')
        if not jnd_itd:
            raise ValueError('jnd_itd not found in previous results file. Please run the full experiment.')
    if do_interference:
        print('The main part of the experiment starts now (interference task).')
        print('Blocks of about 4min each are presented with pauses inbetween.')
        for _ in range(10):
            interference_block(jnd_room, jnd_voice, jnd_itd)

if __name__ == '__main__':
    main_experiment()



'''
Zhenyu's experiment:
Is motion adaptation speed-dependent?
Use like this:
>>> from slab.experiments import motion_speed
>>> motion_speed.main_experiment('subject01')
'''

import time
import functools
import numpy
import scipy
import slab

# configuration
slab.Signal.set_default_samplerate(44100)
_speaker_positions = numpy.arange(-90, 0.01, 4)
_results_file = None
_adapter_speed = 150
_adapter_dir = 'left'
_n_adapters_per_trial = 6
_n_blocks_per_speed = 2
_after_stim_pause = 0.1
_speeds = [50, 100, 150, 200, 250]  # deg/sec
_jnd_diff_thresh = 1.5

slab.ResultsFile.results_folder = 'Results'


def moving_gaussian(speed=100, width=7.5, SNR=10, direction='left'):
    '''
    Make a wide Gauss curve shaped stimulus that moves horizontally across
    a range of virtual loudspeakers. This is the base stimulus of the experiment.
    '''
    if direction == 'left':
        dir = -1
        starting_loc = _speaker_positions[-1]
    else:
        dir = 1
        starting_loc = _speaker_positions[0]

    def loc(time):
        return (speed * time) * dir + starting_loc
    # make times vector from speed and positions angle difference
    end_time = _speaker_positions.ptp() / speed
    time_delta = 0.01  # 10 ms
    times = numpy.arange(0, end_time + time_delta, time_delta)
    # step through time, saving speaker amplitudes for each step
    speaker_amps = numpy.zeros((len(_speaker_positions), len(times)))
    for idx, t in enumerate(times):
        speaker_amps[:, idx] = scipy.stats.norm.pdf(_speaker_positions, loc=loc(t), scale=width)
    # scale the amplitudes to max 0, min -SNR dB
    maximum = scipy.stats.norm.pdf(0, loc=0, scale=width)
    minimum = speaker_amps.min()
    speaker_amps = numpy.interp(speaker_amps, [minimum, maximum], [-SNR, 0])
    speaker_signals = []
    hrtf = slab.HRTF.kemar()
    for i, speaker_position in enumerate(_speaker_positions):
        sig = slab.Binaural.pinknoise(duration=end_time)
        spectral_shape = hrtf.interpolate(azimuth=speaker_position, elevation=0)
        sig = spectral_shape.apply(sig)  # apply HRTF filter from correct direction (ITD still missing)
        itd = sig.azimuth_to_itd(azimuth=speaker_position)
        sig = sig.itd(itd)  # add the necessary ITD
        sig = sig.envelope(apply_envelope=speaker_amps[i, :], times=times, kind='dB')
        speaker_signals.append(sig)
    sig = speaker_signals[0]
    for speaker_signal in speaker_signals[1:]:  # add sounds
        sig += speaker_signal
    sig /= len(_speaker_positions)
    sig.ramp(duration=end_time/3)  # ramp the sum
    sig.filter(frequency=[500, 14000], kind='bp')
    sig.level = 75
    return sig


def familiarization():
    '''
    Presents the familiarization stimuli (100% modulation depth, random direction)
    '''
    print('Familiarization: sounds moving left or right are presented.')
    print('The direction should be easy to hear.')
    print('Press 1 for left, 2 for right.')
    input('Press enter to start familiarization (2min)...')
    repeat = 'r'
    while repeat == 'r':
        trials = slab.Trialsequence(
            conditions=['left', 'right'], n_reps=10, kind='random_permutation')
        responses = []
        _results_file.write('familiarization:', tag='time')
        for dir in trials:
            stim = moving_gaussian(speed=_adapter_speed, SNR=100, direction=dir)
            stim.play()  # present
            with slab.psychoacoustics.key() as key:  # and get response
                resp = key.getch()
            if dir == 'left':  # transform response: left = key '1', right = key '2'
                resp = resp == 49
            else:
                resp = resp == 50
            responses.append(resp)
            time.sleep(_after_stim_pause)
        # compute hitrate
        hitrate = sum(responses)/trials.n_trials
        print(f'hitrate: {hitrate}')
        repeat = input('Press enter to continue, "r" to repeat familiarization.')
    _results_file.write(hitrate, tag='hitrate')
    return hitrate


def practice_stairs():
    '''
    Presents an easy large-step staircase to practice.
    '''
    print('This is a practise run. Explain the procedure to the participant.')
    print('Show the running staircase to the participant.')
    print('')
    print('One sound is presented in each trial.')
    print('Is this sound moving left or right?')
    print('Press 1 for left, 2 for right.')
    print('The direction will get more and more difficult to hear.')
    input('Press enter to start practice...')
    stairs = slab.Staircase(start_val=24, n_reversals=6,
                            step_sizes=[10, 6, 4], min_val=1, max_val=30, n_up=1, n_down=1, n_pretrials=1)
    for trial in stairs:
        direction = numpy.random.choice(('left', 'right'))
        stim = moving_gaussian(speed=_adapter_speed, SNR=trial, direction=direction)
        stairs.present_tone_trial(
            stimulus=stim, correct_key_idx=0 if direction == 'left' else 1, print_info=True)
        stairs.plot()
    input('Done. Press enter to continue...')
    stairs.close_plot()


def jnd(speed=None, adapter_list=None):
    '''
    Presents a staricase of moving_gaussian stimuli with varying SNR and returns the threshold.
    This threshold is used in the main experiment as the listener-specific SNR parameter.
    '''
    if adapter_list:
        print(f'{_n_adapters_per_trial} sounds moving {_adapter_dir}, followed by')
        print('one sound moving left or right is presented.')
        print('Is this last sound moving left or right?')
    else:
        print('One sound is presented in each trial.')
        print('Is this sound moving left or right?')
    print('Press 1 for left, 2 for right.')
    print('The direction will get more and more difficult to hear.')
    input('Press enter to start JND estimation...')
    repeat = 'r'
    while repeat == 'r':
        # define the staircase
        stairs = slab.Staircase(start_val=24, n_reversals=14,
                                step_sizes=[8, 6, 4, 3, 2, 1], min_val=0, max_val=30, n_up=1, n_down=1, n_pretrials=2)
        # loop through it
        _results_file.write('jnd:', tag='time')
        for trial in stairs:
            direction = numpy.random.choice(('left', 'right'))
            stim = moving_gaussian(speed=speed, SNR=trial, direction=direction)
            if adapter_list:
                adapters = adapter_list.random_choice(
                    n=_n_adapters_per_trial)  # some variety in the adapters
                adapters.append(stim)  # add stim to list of adapters
                stim = slab.Sound.sequence(*adapters)  # concatenate sounds in the list
            stairs.present_tone_trial(
                stimulus=stim, correct_key_idx=0 if direction == 'left' else 1)
        thresh = stairs.threshold(n=10)
        tag = f"{speed} {'with_adapter' if adapter_list else 'no_adapter'}"
        print(f'jnd for {tag}: {round(thresh, ndigits=1)}')
        repeat = input('Press enter to continue, "r" to repeat this threshold measurement.\n\n')
        _results_file.write(thresh, tag=tag)
    return thresh


def make_adapters():
    'Pre-make many adapter instances to speed-up constructing the stimuli.'
    kwargs = {'speed': _adapter_speed, 'SNR': 100, 'direction': _adapter_dir}
    make_adapter = functools.partial(moving_gaussian, **kwargs)
    adapter_list = slab.Precomputed(make_adapter, _n_adapters_per_trial * 3)
    return adapter_list


def main_experiment(subject=None):
    '''
    A complex spatially extended moving sound is generated ('moving_gaussian'). This stimulus simulates the acoustics of a free-field loudspeaker arc.  A gaussian profile moves from left to right or right to left across the virtual speaker array and the speed of the movement and modulation depth (across space) can be varied. Detection thresholds for motion direction are measured at different motion speeds. Then the effect of adaptation by a long moving adapter at one speed on the detectability of motion at different speeds is measured.

    This experiment showcases complex stimulus generation and staircases, among others.
    '''
    global _results_file
    # set up the results file
    if not subject:
        subject = input('Enter subject code: ')
    _results_file = slab.ResultsFile(subject=subject)
    _ = familiarization() # run the familiarization, the hitrate is saved in the results file
    practice_stairs()  # run the stairs practice
    print('The main part of the experiment starts now (motion direction thresholds).')
    adapter_list = make_adapters()
    # results table with three colums: speed, jnd_no_adapter, jnd_adapter
    jnds = numpy.zeros((len(_speeds), 3, _n_blocks_per_speed))
    repeats = list()  # for collecting conditions marked for re-measuring due to large differences in jnds
    for i in range(_n_blocks_per_speed):  # if larger t
        speed_seq = slab.Trialsequence(conditions=_speeds, n_reps=1, kind='random_permutation')
        for speed in speed_seq:
            idx = _speeds.index(speed)  # index of the current speed value, used for results table
            jnds[idx, 0, :] = speed
            if numpy.random.choice((True, False)):  # presented without adapters first
                # each call to jnd prints instructions and saves to the results file
                jnds[idx, 1, i] = jnd(speed)
                jnds[idx, 2, i] = jnd(speed, adapter_list)
            else:  # with adapters first
                jnds[idx, 2, i] = jnd(speed, adapter_list)
                jnds[idx, 1, i] = jnd(speed)
            if i == 1:
                # if measurements too different, mark for repeat
                if abs(jnds[idx, 1, 1] - jnds[idx, 1, 0]) > _jnd_diff_thresh:
                    repeats.append((speed, False))
                    print(f'{speed} no adapter: Difference to first JND too large. Marked for repetition!')
                # if measurements too different, mark for repeat
                if abs(jnds[idx, 2, 1] - jnds[idx, 2, 0]) > _jnd_diff_thresh:
                    repeats.append((speed, True))
                    print(f'{speed} with adapters: Difference to first JND too large. Marked for repetition!')
        # save a string representation of the numpy results table
        _results_file.write(str(jnds[:, :, i]), tag=f'results round {i}')
    # rerun the conditions marked for repeated measurements
    if repeats:
        print('Repeating marked JND measurements.')
        for speed, adapters in repeats:  # these will just be tagged in the results file, not in the tables
            if adapters:
                jnd(speed, adapter_list)
            else:
                jnd(speed)


if __name__ == '__main__':
    main_experiment(subject='test')



import inspect
import slab
import numpy
from copy import deepcopy

ils = slab.Binaural.make_interaural_level_spectrum()
hrtf = slab.HRTF.kemar()

def test_itd():
    for _ in range(100):
        sound = slab.Binaural.whitenoise()
        itd = numpy.random.uniform(-0.1, 0.1)
        lateral = sound.itd(duration=itd)
        itd = slab.Sound.in_samples(itd, 8000)
        assert itd == lateral.itd(max_lag=numpy.abs(itd))


def test_ild():
    sound = slab.Binaural.whitenoise()
    for _ in range(100):
        ild = numpy.random.uniform(-10, 10)
        lateral = sound.ild(ild)
        numpy.testing.assert_almost_equal(lateral.ild(), ild, decimal=5)


def test_azimuth_to_itd_ild():
    for _ in range(100):
        frequency = numpy.random.randint(50, 2000)
        headsize = numpy.random.uniform(7, 11)
        itds = []
        ilds = []
        for azimuth in numpy.linspace(-90, 0, 20):
            itds.append(slab.Binaural.azimuth_to_itd(azimuth, frequency, headsize))
            ilds.append(slab.Binaural.azimuth_to_ild(azimuth, frequency, ils))
        assert all([itds[i] < itds[i+1] for i in range(len(itds)-1)])
        itds = []
        for azimuth in numpy.linspace(90, 0, 20):
            itds.append(slab.Binaural.azimuth_to_itd(azimuth, frequency, headsize))
            ilds.append(slab.Binaural.azimuth_to_ild(azimuth, frequency, ils))
            assert all([itds[i] > itds[i + 1] for i in range(len(itds) - 1)])


def test_at_azimuth():
    for _ in range(10):
        sound = slab.Binaural.whitenoise()
        for azimuth in numpy.linspace(-90, 90, 40):
            lateral = sound.at_azimuth(azimuth, ils=ils)
            itd = slab.Sound.in_samples(slab.Binaural.azimuth_to_itd(azimuth), 8000)
            assert numpy.abs(itd - lateral.itd()) <= 1
            ild = slab.Binaural.azimuth_to_ild(azimuth, ils=ils)
            assert numpy.abs(ild - lateral.ild()) <= 3


def test_itd_ramp():
    for _ in range(10):
        sound = slab.Binaural.whitenoise()
        from_itd = numpy.random.uniform(-0.001, 0.)
        to_itd = numpy.random.uniform(0.001, 0.)
        moving = sound.itd_ramp(from_itd=from_itd, to_itd=to_itd)
        start, stop = deepcopy(moving), deepcopy(moving)
        start.data = start.data[0:200, :]
        stop.data = stop.data[-201:-1, :]
        assert numpy.abs(start.itd() - slab.Sound.in_samples(from_itd, 8000)) <= 1
        assert numpy.abs(stop.itd() - slab.Sound.in_samples(to_itd, 8000)) <= 1


def test_ild_ramp():
    for _ in range(100):
        sound = slab.Binaural.whitenoise()
        from_ild = numpy.random.uniform(-10, 0.)
        to_ild = numpy.random.uniform(10, 0.)
        moving = sound.ild_ramp(from_ild=from_ild, to_ild=to_ild)
        start, stop = deepcopy(moving), deepcopy(moving)
        start.data = start.data[0:200, :]
        stop.data = stop.data[-201:-1, :]
        numpy.testing.assert_almost_equal(numpy.diff(start.level), from_ild, decimal=0)
        numpy.testing.assert_almost_equal(numpy.diff(stop.level), to_ild, decimal=0)


def test_externalize():
    for _ in range(10):
        idx_frontal = numpy.where((hrtf.sources.vertical_polar[:, 1] == 0) &
                                  (hrtf.sources.vertical_polar[:, 0] == 0))[0][0]
        sound = slab.Binaural.whitenoise(samplerate=hrtf.samplerate)
        filtered = hrtf.data[idx_frontal].apply(sound)
        filtered = filtered.resize(sound.duration)
        external = sound.externalize()
        assert numpy.abs(filtered.data-external.data).sum() < numpy.abs(filtered.data-sound.data).sum()
        assert numpy.abs(sound.level - external.level).max() < 0.6


def test_interaural_level_spectrum():
    sound = slab.Binaural.whitenoise(samplerate=int(ils['samplerate']))
    azimuths = ils['azimuths']
    for i, azi in enumerate(azimuths):
        level_differences = ils['level_diffs'][:, i]
        lateral = sound.interaural_level_spectrum(azi, ils)
        fbank = slab.Filter.cos_filterbank(samplerate=sound.samplerate, pass_bands=True)
        subbands_left = fbank.apply(lateral.left)
        subbands_right = fbank.apply(lateral.right)
        assert -1 < (level_differences - (subbands_left.level - subbands_right.level)).mean() < 1


def test_overloaded_sound_generators():
    methods = ['chirp', 'click', 'clicktrain', 'dynamic_tone', 'equally_masking_noise',
               'harmoniccomplex', 'irn', 'multitone_masker', 'pinknoise', 'powerlawnoise',
               'silence', 'tone', 'vowel', 'whitenoise']
    for method in methods:
        func = getattr(slab.Binaural, method)
        assert func().n_channels == 2
        assert func().name != 'unnamed'


def test_drr():
    for _ in range(10):
        steepness = numpy.random.randint(1, 3)
        duration = numpy.random.uniform(2.0, 8.0)
        decay_resolution = 10
        decay_curve = [numpy.exp(-i * steepness) for i in range(decay_resolution)]
        sound = slab.Binaural.whitenoise(kind='dichotic', duration=duration, samplerate=44100)
        impulse = sound.envelope(apply_envelope=decay_curve)
        winlength = numpy.random.uniform(0.0001, 0.01)
        if numpy.random.randint(0, 2):
            winlength = max(2, impulse.in_samples(winlength, impulse.samplerate))
        if numpy.random.randint(0, 2):
            drr = impulse.drr(winlength=winlength)
        else:
            drr = impulse.drr()
        assert 0 > drr > -100
        assert isinstance(drr, float)



import copy
import numpy

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = False
try:
    import scipy.signal
except ImportError:
    scipy = False

import slab.signal
from slab.signal import Signal  # getting the base class


class Filter(Signal):
    """
    Class for generating and manipulating filter banks and transfer functions. Filters can be finite impulse
    response ('FIR'), impulse response ('IR'), or transfer functions ('TF'). FIR filters are applied using
    two-way filtering (see scipy.signal.filtfilt), which avoids adding group delays and is intended for high-
    and lowpass filters and the like. IR filters are applied using convolution (scipy.signal.fftconvolve) and
    are intended for long room impulse responses, binaural impulse responses and the like, where group delays
    are intended.

    Arguments:
        data (numpy.ndarray | slab.Signal | list): samples of the filter. If it is an array, the first dimension
            should represent the number of samples and the second one the number of channels. If it's an object,
            it must have a .data attribute containing an array. If it's a list, the elements can be arrays or objects.
            The output will be a multi-channel sound with each channel corresponding to an element of the list.
        samplerate (int | None): the samplerate of the sound. If None, use the default samplerate.
        fir (str): the kind of filter; options are 'FIR', 'IR', or 'TF'.
    Attributes:
        .n_filters: number of filters in the object (overloads `n_channels` attribute in the parent `Signal` class)
        .n_taps: the number of taps in a finite impulse response filter. Analogous to `n_samples` in a `Signal`.
        .n_frequencies: the number of frequency bins in a Fourier filter. Analogous to `n_samples` in a `Signal`.
        .frequencies: the frequency axis of a Fourier filter.
    """
    # instance properties
    n_filters = property(fget=lambda self: self.n_channels, doc='The number of filters in the bank.')
    n_taps = property(fget=lambda self: self.n_samples, doc='The number of filter taps.')
    n_frequencies = property(fget=lambda self: self.n_samples, doc='The number of frequency bins.')
    frequencies = property(fget=lambda self: numpy.fft.rfftfreq(self.n_frequencies * 2 - 1, d=1 / self.samplerate)
                           if self.fir=='TF' else None, doc='The frequency axis of the filter.')

    def __init__(self, data, samplerate=None, fir='FIR'):
        if isinstance(fir, bool):
            raise AttributeError("Assigning a boolean to fir is deprecated. Use 'TF' instead of False, and 'FIR' or 'IR' otherwise.")
        if 'IR' in fir and scipy is False:
            raise ImportError('(F)IR filters require scipy.')
        super().__init__(data, samplerate)
        if not (fir=='FIR' or fir=='IR' or fir=='TF'):
            raise ValueError("fir must be 'FIR', 'IR', or 'TF'!")
        self.fir = fir

    def __repr__(self):
        return f'{type(self)} (\n{repr(self.data)}\n{repr(self.samplerate)}\n{repr(self.fir)})'

    def __str__(self):
        if 'IR' in self.fir:
            return f'{type(self)}, filters {self.n_filters}, FIR: taps {self.n_taps}, samplerate {self.samplerate}'
        return f'{type(self)}, filters {self.n_filters}, TF: freqs {self.n_frequencies}, samplerate {self.samplerate}'

    @staticmethod
    def band(kind='hp', frequency=100, gain=None, samplerate=None, length=1000, fir='FIR'):
        """
        Generate simple passband or stopband filters, or filters with a transfer function defined by  pairs
        of `frequency` and `gain` values.

        Arguments:
            kind: The type of filter to generate. Can be 'lp' (lowpass), 'hp' (highpass), 'bp' (bandpass)
                or 'bs' (bandstop/notch). If `gain` is specified the `kind` argument is ignored.
            frequency (int | float | tuple | list): For a low- or highpass filter, a single integer or float value must
                be given which is the filters edge frequency in Hz. a bandpass or -stop filter takes a tuple of two
                values which are the filters lower and upper edge frequencies. Given a list of values and a list of
                equal length as `gain` the resulting filter will have the specified gain at each frequency.
            gain (None | list): Must be None when generating an lowpass, highpass, bandpass or bandstop filter. For
                generating a custom filter, define a list of the same length as `frequency` with values between
                1.0 (no suppression at that frequency) and 0.0 (maximal suppression at that frequency).
            samplerate (int | None): the samplerate of the sound. If None, use the default samplerate.
            length (int): The number of samples in the filter
            fir (str): If 'FIR' or 'IR' generate a finite impulse response filter, else generate a transfer function.
        Returns:
            (slab.Filter): a filter with the specified properties
        Examples::

            filt = slab.Filter.band(frequency=3000, kind='lp')  # lowpass filter
            filt = slab.Filter.band(frequency=(100, 2000), kind='bs')  # bandstop filter
            filt = slab.Filter.band(frequency=[100, 1000, 3000, 6000], gain=[0., 1., 0., 1.])  # custom filter
        """
        if samplerate is None:
            samplerate = slab.signal._default_samplerate
        if 'IR' in fir:  # design a FIR filter
            if scipy is False:
                raise ImportError('Generating FIR filters requires Scipy.')
            if gain is None:  # design band filter
                if kind in ['lp', 'bs']:
                    pass_zero = True
                elif kind in ['hp', 'bp']:
                    pass_zero = False
                if kind in ['hp', 'bs'] and not length % 2:  # high at nyquist requires odd n_taps
                    filt = scipy.signal.firwin(length-1, frequency, pass_zero=pass_zero, fs=samplerate)
                    filt = numpy.append(filt, 0)  # extend to original length
                else:
                    filt = scipy.signal.firwin(length, frequency, pass_zero=pass_zero, fs=samplerate)
            else:
                if not (isinstance(gain, list) and isinstance(frequency, list)):
                    raise ValueError("Gain and frequency must be lists for generating custom filters!")
                dc = dc_gain = []
                if frequency[0] != 0:
                    dc = dc_gain = [0]
                nyq = nyq_gain = []
                if frequency[-1] != samplerate/2:
                    nyq = [samplerate/2]
                    nyq_gain = [0]
                filt = scipy.signal.firwin2(numtaps=length, freq=dc+frequency+nyq, gain=dc_gain+gain+nyq_gain,
                                            fs=samplerate)
        else:  # TF filter
            df = (samplerate/2) / (length-1)
            if gain is None:
                filt = numpy.zeros(length)
                if kind == 'lp':
                    filt[:round(frequency/df)] = 1
                elif kind == 'hp':
                    filt[round(frequency/df):] = 1
                elif kind == 'bp':
                    filt[round(frequency[0]/df):round(frequency[1]/df)] = 1
                elif kind == 'bs':
                    filt[:round(frequency[0]/df)] = 1
                    filt[round(frequency[1]/df):] = 1
                else:
                    raise ValueError('Unknown filter kind. Use lp, hp, bp, or bs.')
            else:
                freqs = numpy.arange(0, (samplerate/2)+df, df)  # frequency bins
                filt = numpy.interp(freqs, numpy.array([0] + frequency + [samplerate/2]), numpy.array([0] + gain + [0]))
        return Filter(data=filt, samplerate=samplerate, fir=fir)

    def apply(self, sig):
        """
        Apply the filter to a sound. If sound and filter have the same number of channels,
        each filter channel will be applied to the corresponding channel in the sound.
        If the filter has multiple channels and the sound only 1, each filter is applied to the same sound.
        In that case the filtered sound wil contain the same number of channels as the filter with every
        channel being a copy of the original sound with one filter channel applied. If the filter has only
        one channel and the sound has multiple channels, the same filter is applied to each sound channel.
        FIR filters are applied using two-way filtering (see scipy.signal.filtfilt), which avoids adding
        group delays and is intended for high- and lowpass filters and the like. IR filters are applied
        using convolution (scipy.signal.fftconvolve) and are intended for long room impulse responses,
        binaural impulse responses and the like, where group delays are intended.

        Arguments:
            sig (slab.Signal | slab.Sound): The sound to be filtered.
        Returns:
            (slab.Signal | slab.Sound): a filtered copy of `sig`.
        Examples::

            filt = slab.Filter.band(frequency=(100, 1500), kind='bp')  # bandpass filter
            sound = slab.Sound.whitenoise()  # generate sound
            filtered_sound = filt.apply(sound)  # apply the filter to the sound
        """
        if (self.samplerate != sig.samplerate) and (self.samplerate != 1):
            raise ValueError('Filter and sound have different sampling rates.')
        out = copy.deepcopy(sig)
        if self.fir == 'FIR':
            if scipy is False:
                raise ImportError('Applying FIR filters requires Scipy.')
            if out.n_samples < self.n_samples * 3:
                padlen = out.n_samples - 1
            else:
                padlen = None
            if self.n_filters == sig.n_channels:  # filter each channel with corresponding filter
                for i in range(self.n_filters):
                    out.data[:, i] = scipy.signal.filtfilt(
                        self.data[:, i], [1], out.data[:, i], axis=0, padlen=padlen)
            elif (self.n_filters == 1) and (sig.n_channels > 1):  # filter each channel
                for i in range(sig.n_channels):
                    out.data[:, i] = scipy.signal.filtfilt(
                        self.data.flatten(), [1], out.data[:, i], axis=0,  padlen=padlen)
            elif (self.n_filters > 1) and (sig.n_channels == 1):  # apply all filters in bank to sound
                out.data = numpy.empty((sig.n_samples, self.n_filters))
                for filt in range(self.n_filters):
                    out.data[:, filt] = scipy.signal.filtfilt(
                        self[:, filt], [1], sig.data, axis=0,  padlen=padlen).flatten()
            else:
                raise ValueError(
                    'Number of filters must equal number of sound channels, or either one of them must be equal to 1.')
        elif self.fir == 'IR':
            length = sig.n_samples + self.n_taps - 1
            if self.n_filters == sig.n_channels:  # filter each channel with corresponding filter
                out.data = numpy.empty((length, sig.n_channels))
                for i in range(self.n_filters):
                    out.data[:, i] = scipy.signal.fftconvolve(sig[:, i], self[:, i], mode='full')
            elif (self.n_filters == 1) and (sig.n_channels > 1):  # filter each channel
                out.data = numpy.empty((length, sig.n_channels))
                for i in range(sig.n_channels):
                    out.data[:, i] = scipy.signal.fftconvolve(sig[:, i], self[:, 0], mode='full')
            elif (self.n_filters > 1) and (sig.n_channels == 1):  # apply all filters in bank to sound
                out.data = numpy.empty((length, self.n_filters))
                for filt in range(self.n_filters):
                    out.data[:, filt] = scipy.signal.fftconvolve(sig[:, 0], self[:, filt], mode='full').flatten()
            else:
                raise ValueError(
                    'Number of filters must equal number of sound channels, or either one of them must be equal to 1.')
        else:  # FFT filter
            sig_rfft = numpy.fft.rfft(sig.data, axis=0)
            sig_freq_bins = numpy.fft.rfftfreq(sig.n_samples, d=1 / sig.samplerate)
            filt_freq_bins = self.frequencies
            # interpolate the FFT filter bins to match the length of the fft of the sound
            if self.n_filters == sig.n_channels:  # filter each channel with corresponding filter
                for chan in range(sig.n_channels):
                    _filt = numpy.interp(sig_freq_bins, filt_freq_bins, self[:, chan])
                    out.data[:, chan] = numpy.fft.irfft(sig_rfft[:, chan] * _filt, sig.n_samples)
            elif (self.n_filters == 1) and (sig.n_channels > 1):  # filter each channel
                _filt = numpy.interp(sig_freq_bins, filt_freq_bins, self.data.flatten())
                for chan in range(sig.n_channels):
                    out.data[:, chan] = numpy.fft.irfft(sig_rfft[:, chan] * _filt, sig.n_samples)
            elif (self.n_filters > 1) and (sig.n_channels == 1):  # apply all filters in bank to sound
                out.data = numpy.empty((sig.n_samples, self.n_filters))
                for filt in range(self.n_filters):
                    _filt = numpy.interp(sig_freq_bins, filt_freq_bins, self[:, filt])
                    out.data[:, filt] = numpy.fft.irfft(sig_rfft.flatten() * _filt, sig.n_samples)
            else:
                raise ValueError(
                    'Number of filters must equal number of sound channels, or either one of them must be equal to 1.')
        return out

    def tf(self, channels='all', n_bins=None, show=True, axis=None):
        """
        Compute a filter's transfer function (magnitude over frequency) and optionally plot it.

        Arguments:
            channels (str | list | int): the filter channels to compute the transfer function for. Defaults to the
                string "all" which includes all channels. To compute the transfer function for multiple channels, pass
                a list of channel integers. For the transfer function for a single channel pass it's index as integer.
            n_bins (None): number of bins in the transfer function (determines frequency resolution).
                If None, use the maximum number of bins.
            show (bool): whether to show the plot right after drawing.
            axis (matplotlib.axes.Axes | None): axis to plot to. If None create a new plot.
        Returns:
            (numpy.ndarray): the frequency bins in the range from 0 Hz to the Nyquist frequency.
            (numpy.ndarray: the magnitude of each frequency in `w`.
            None: If `show` is True OR and `axis` was specified, a plot is drawn and nothing is returned.
        Examples::

            filt = slab.Filter.band(frequency=(100, 1500), kind='bp')  # bandpass filter
            filt.tf(show=True)  # compute and plot the transfer functions
            w, h = filt.tf(show=False)  # compute and return the transfer functions
        """
        if isinstance(channels, int):
            if channels > self.n_filters:  # check chan is in range of n_filters
                raise IndexError("Channel index out of range!")
            else:
                channels = [channels]
        elif isinstance(channels, list):  # check that all the elements are unique and in range
            if len(channels) != len(set(channels)):
                raise ValueError("There should be no duplicates in the list of channels!")
            if min(channels) < 0 or max(channels) < self.n_filters:
                raise IndexError("Channel index out of range!")
            if not all([isinstance(i, int) for i in channels]):
                raise ValueError("Channels must be integers!")
        elif channels == 'all':
            channels = list(range(self.n_filters))  # now we have a list of filter indices to process
        if n_bins is None:
            n_bins = self.data.shape[0]
        if 'IR' in self.fir:
            if scipy is False:
                raise ImportError('Computing transfer functions of FIR filters requires Scipy.')
            h = numpy.empty((n_bins, len(channels)))
            w = numpy.empty(n_bins)
            for i, idx in enumerate(channels):
                w, _h = scipy.signal.freqz(self.channel(idx), worN=n_bins, fs=self.samplerate)
                h[:, i] = 20 * numpy.log10(numpy.abs(_h.flatten()))
        else:
            w = self.frequencies
            data = self.data[:, channels]
            data[data == 0] += numpy.finfo(float).eps
            h = 20 * numpy.log10(data)
            if not n_bins == len(w):  # interpolate if necessary
                w_interp = numpy.linspace(0, w[-1], n_bins)
                h_interp = numpy.zeros((n_bins, len(channels)))
                for idx, _ in enumerate(channels):
                    h_interp[:, idx] = numpy.interp(w_interp, w, h[:, idx])
                h = h_interp
                w = w_interp
        if show or (axis is not None):
            if plt is False:
                raise ImportError('Plotting transfer functions requires matplotlib.')
            if axis is None:
                _, axis = plt.subplots()
            axis.plot(w, h)
            axis.set(title='Frequency [Hz]', ylabel='Amplitude [dB]', xlabel='Frequency Response')
            axis.grid(True)
            if show:
                plt.show()
        else:
            return w, h

    @staticmethod
    # TODO: oversampling factor needed for cochleagram!
    def cos_filterbank(length=5000, bandwidth=1/3, low_cutoff=0, high_cutoff=None, pass_bands=False, n_filters=None, samplerate=None):
        """
        Generate a set of Fourier filters. Each filter's transfer function is given by the positive phase of a
        cosine wave. The amplitude of the cosine is that filters central frequency. Following the organization of the
        cochlea, the width of the filter increases in proportion to it's center frequency. This increase is defined
        by Moore & Glasberg's formula for the equivalent rectangular bandwidth (ERB) of auditory filters.
        The number of filters is either determined by the `n_filters` argument or calculated based on the desired
        `bandwidth` or. This function is used for example to divide a sound into sub-bands for equalization.

        Attributes:
            length (int): The number of bins in each filter, determines the frequency resolution.
            bandwidth (float): Width of the sub-filters in octaves. The smaller the bandwidth, the more filters will be generated.
            low_cutoff (int | float): The lower limit of frequency range in Hz.
            high_cutoff (int | float): The upper limit of frequency range in Hz. If None, use the Nyquist frequency.
            pass_bands (bool): Whether to include a half cosine at the filter bank's lower and upper edge frequency.
                If True, allows reconstruction of original bandwidth when collapsing subbands.
            n_filters (int | None): Number of filters. When this is not None, the `bandwidth` argument is ignored.
            samplerate (int | None): the samplerate of the sound that the filter shall be applied to.
                If None, use the default samplerate.s
        Examples::

            sig = slab.Sound.pinknoise(samplerate=44100)
            fbank = slab.Filter.cos_filterbank(length=sig.n_samples, bandwidth=1/10, low_cutoff=100,
                                          samplerate=sig.samplerate)
            fbank.tf()
            # apply the filter bank to the data. The filtered sound will contain as many channels as there are
            # filters in the bank. Every channel is a copy of the original sound with one filter applied.
            # In this context, the channels are the signals sub-bands:
            sig_filt = fbank.apply(sig)
        """
        if samplerate is None:
            samplerate = slab.signal._default_samplerate
        if not high_cutoff:
            high_cutoff = samplerate / 2
        freq_bins = numpy.fft.rfftfreq(length, d=1/samplerate)
        n_freqs = len(freq_bins)
        center_freqs, bandwidth, erb_spacing = Filter._center_freqs(
            low_cutoff=low_cutoff, high_cutoff=high_cutoff, bandwidth=bandwidth, pass_bands=pass_bands, n_filters=n_filters)
        n_filters = len(center_freqs)
        filts = numpy.zeros((n_freqs, n_filters))
        freqs_erb = Filter._freq2erb(freq_bins)
        for i in range(n_filters):
            l = center_freqs[i] - erb_spacing
            h = center_freqs[i] + erb_spacing
            avg = center_freqs[i]  # center of filter
            width = erb_spacing * 2  # width of filter
            filts[(freqs_erb > l) & (freqs_erb < h), i] = numpy.cos(
                (freqs_erb[(freqs_erb > l) & (freqs_erb < h)] - avg) / width * numpy.pi)
        return Filter(data=filts, samplerate=samplerate, fir='TF')

    @staticmethod
    def _center_freqs(low_cutoff, high_cutoff, bandwidth=1/3, pass_bands=False, n_filters=None):
        ref_freq = 1000  # Hz, reference for conversion between oct and erb bandwidth
        h = Filter._freq2erb(high_cutoff)
        l = Filter._freq2erb(low_cutoff)
        ref_erb = Filter._freq2erb(ref_freq)
        if n_filters is None:
            erb_spacing = Filter._freq2erb(ref_freq*2**bandwidth) - ref_erb
            n_filters = int(numpy.round((h - l) / erb_spacing))
        elif n_filters is not None and pass_bands is False:
            # add 2 so that after omitting pass_bands we get the desired n_filt
            n_filters += 2
        center_freqs, erb_spacing = numpy.linspace(l, h, n_filters, retstep=True)
        if not pass_bands:
            center_freqs = center_freqs[1:-1]  # exclude low and highpass filters
        bandwidth = numpy.log2(Filter._erb2freq(ref_erb + erb_spacing) /
                               ref_freq)  # convert erb_spacing to octaves
        return center_freqs, bandwidth, erb_spacing

    @staticmethod
    def collapse_subbands(subbands, filter_bank=None):
        """
        Sum a sound that has been filtered with a filterbank and which channels represent the sub-bands of
        the original sound. For each sound channel, the fourier transform is calculated and the result is
        multiplied with the corresponding filter in the filter bank. For the resulting spectrum, and inverse
        fourier transform is performed. The resulting sound is summed over the all channels.

        Arguments:
            subbands (slab.Signal): The sound which is divided into subbands by filtering. The number of channels
                in the sound must be equal to the number of filters in the filter bank.
            filter_bank (None | slab.Filter): The filter bank applied to the sound's subbands. The number of
                filters must be equal to the number of channels in the sound. If None a filter bank with the default
                parameters is generated. Note that the filters must have a number of frequency bins equal to the
                number of samples in the sound.
        Returns:
            (slab.Signal): A sound generated from summing the spectra of the subbands.
        Examples::

            sig = slab.Sound.whitenoise()  # generate a sound
            fbank = slab.Filter.cos_filterbank(length=sig.n_samples)  # generate a filter bank
            subbands = fbank.apply(sig)  # divide the sound into subbands by applying the filter
            # by collapsing the subbands, a new sound is generated that is (almost) equal to the original sound:
            collapsed = fbank.collapse_subbands(subbands, fbank)
        """
        new = copy.deepcopy(subbands)
        if not filter_bank:
            filter_bank = Filter.cos_filterbank(
                length=subbands.n_samples, samplerate=subbands.samplerate)
        elif 'IR' in filter_bank.fir:
            raise ValueError("Not implemented for FIR filters!")
        if subbands.samplerate != filter_bank.samplerate:
            raise ValueError('Signal and filter bank need to have the same samplerate!')
        subbands_rfft = numpy.fft.rfft(subbands.data, axis=0)
        subbands = numpy.fft.irfft(subbands_rfft * filter_bank.data, axis=0)
        collapsed = Signal(data=subbands.sum(axis=1), samplerate=filter_bank.samplerate)
        new.data, new.samplerate = collapsed.data, collapsed.samplerate
        return new

    def filter_bank_center_freqs(self):
        """
        Get the maximum of each filter in a filter bank. For filter banks generated with the `cos_filterbank`
        method this corresponds to the filters center frequency.

        Returns:
            (numpy.ndarray): array with length equal to the number of filters in the bank, containing each filter's
                center frequency.
        """
        if 'IR' in self.fir:
            raise NotImplementedError('Not implemented for FIR filter banks.')
        freqs = self.frequencies
        center_freqs = numpy.zeros(self.n_filters)
        for i in range(self.n_filters):  # for each filter
            idx = numpy.argmax(self.channel(i).data)  # get index of max Gain
            center_freqs[i] = freqs[idx]  # look-up freqe of index -> centre_freq for that filter
        return center_freqs

    @staticmethod
    def equalizing_filterbank(reference, sound, length=1000, bandwidth=1/8, low_cutoff=200, high_cutoff=None,
                              alpha=1.0, filt_meth='filtfilt'):
        """
        Generate an equalizing filter from the spectral difference between a `sound` and a `reference`. Both are
        divided into sub-bands using the `cos_filterbank` and the level difference per sub-band is calculated. The
        sub-band frequencies and level differences are then used to generate an equalizing filter that makes the
        spectrum of the `sound` more equal to the one of the `reference`. The main use case is equalizing the
        differences between transfer functions of individual loudspeakers.

        Arguments:
            reference (slab.Sound): The reference for equalization, i.e. what the sound should look like after
                applying the equalization. Must have exactly one channel.
            sound (slab.Sound): The sound to equalize. Can have multiple channels.
            length (int): Number of frequency bins in the filter.
            bandwidth (float): Width of the filters, used to divide the signal into subbands, in octaves. A small
                bandwidth results in a fine tuned transfer function which is useful for equalizing small notches.
            low_cutoff (int | float): The lower limit of frequency range in Hz.
            high_cutoff (int | float): The upper limit of frequency range in Hz. If None, use the Nyquist frequency.
            alpha (float): Filter regularization parameter. Values below 1.0 reduce the filter's effect, values above
                amplify it. WARNING: large filter gains may result in temporal distortions of the sound
            filt_meth (str): filtering method, must be in ('filtfilt', 'fft', 'slab', 'causal'). 'filtfilt' and 'slab'
                applies the filter twice (forward and backward) so the filtered signal is not time-shifted, but in this
                case we need a filter with half of the desired amplitude so the end result has the correct modification.
                For 'fft' and 'causal', the filter with desired amplitude is returned
        Returns:
            (slab.Filter): An equalizing filter bank with a number of filters equal to the number of channels in the
                equalized `sound`.
        Example::

            # generate a sound and apply some arbitrary filter to it
            sound = slab.Sound.pinknoise(samplerate=44100)
            filt = slab.Filter.band(frequency=[100., 800., 2000., 4300., 8000., 14500., 18000.],
                                    gain=[0., 1., 0., 1., 0., 1., 0.], samplerate=sound.samplerate)
            filtered = filt.apply(sound)
            # make an equalizing filter and apply it to the filtered signal. The result looks more like the original
            fbank = slab.Filter.equalizing_filterbank(sound, filtered, low_cutoff=200, high_cutoff=16000)
            equalized = fbank.apply(filtered)
        """
        if scipy is False:
            raise ImportError('Generating equalizing filter banks requires Scipy.')
        if filt_meth not in ('filtfilt', 'fft', 'slab', 'causal'):
            raise ValueError("filt_meth must be one of {'filtfilt', 'fft', 'slab', 'causal'}")
        if reference.n_channels > 1:
            raise ValueError("The reference sound must have only one channel!")
        if bool(reference.n_samples % 2):  # number of samples must be even:
            reference.resize(reference.n_samples + 1)
        if bool(sound.n_samples % 2):
            sound.resize(sound.n_samples + 1)
        if reference.samplerate > sound.samplerate:  # resample higher to lower rate if necessary
            reference = reference.resample(sound.samplerate)
        elif reference.samplerate < sound.samplerate:
            sound = sound.resample(reference.samplerate)
        if high_cutoff is None:
            high_cutoff = reference.samplerate / 2
        # different conditions on filtering method
        is_fir = 'FIR'
        filt_scaling = 1
        if filt_meth == 'fft':
            is_fir = 'TF'
        elif filt_meth in ('slab', 'filtfilt'):
            filt_scaling = 0.5

        fbank = Filter.cos_filterbank(length=length, bandwidth=bandwidth,
                                      low_cutoff=low_cutoff, high_cutoff=high_cutoff, samplerate=reference.samplerate)
        center_freqs, _, _ = Filter._center_freqs(low_cutoff, high_cutoff, bandwidth)
        center_freqs = list(Filter._erb2freq(center_freqs))
        center_freqs[-1] = min(center_freqs[-1], reference.samplerate / 2)
        # level of the reference in each of the subbands
        levels_target = fbank.apply(reference).level
        # make it the same shape as levels_signal
        levels_target = numpy.tile(levels_target, (sound.n_channels, 1)).T
        # level of each channel in the sound in each of the subbands
        levels_signal = numpy.ones((len(center_freqs), sound.n_channels))
        for idx in range(sound.n_channels):
            levels_signal[:, idx] = fbank.apply(sound.channel(idx)).level
        amp_diffs_dB = levels_target - levels_signal
        # if we want to keep the scaling
        amp_diffs_dB = filt_scaling * amp_diffs_dB * alpha
        # convert differences from dB to ratio
        amp_diffs = 10 ** (amp_diffs_dB / 20.)
        center_freqs = [0] + center_freqs
        filts = numpy.zeros((length, sound.n_channels))
        for chan in range(sound.n_channels):
            gain = [1] + list(amp_diffs[:, chan])
            filt = Filter.band(frequency=list(center_freqs), gain=gain, samplerate=reference.samplerate, fir=is_fir,
                               length=length)
            filts[:, chan] = filt.data.flatten()
        return Filter(data=filts, samplerate=reference.samplerate, fir=is_fir)

    def save(self, filename):
        """
        Save the filter in Numpy's .npy format to a file.
        Arguments:
            filename(str | pathlib.Path): Full path to which the data is saved.
        """
        fs = numpy.tile(self.samplerate, reps=self.n_filters)
        if self.fir == 'IR':
            fir_coded = 1
        elif self.fir == 'TF':
            fir_coded = 2
        else:
            fir_coded = 0
        fir = numpy.tile(fir_coded, reps=self.n_filters)
        fs = fs[numpy.newaxis, :]
        fir = fir[numpy.newaxis, :]
        to_save = numpy.concatenate((fs, fir, self.data))  # prepend the samplerate as new 'filter'
        numpy.save(filename, to_save)

    @staticmethod
    def load(filename):
        """
        Load a filter from a .npy file.

        Arguments:
            filename(str | pathlib.Path): Full path to the file to load.
        Returns:
            (slab.Filter): The `Filter` loaded from the file.
        """
        data = numpy.load(filename)
        samplerate = data[0][0]  # samplerate is in the first filter
        fir_coded = data[1][0]  # fir is in the first filter
        if fir_coded == 2:
            fir = 'TF'
        elif fir_coded == 1:
            fir = 'IR'
        else:
            fir = 'FIR'
        data = data[2:, :]  # drop the samplerate and fir entries
        return Filter(data, samplerate=samplerate, fir=fir)

    @staticmethod
    def _freq2erb(freq_hz):
        """ Converts Hz to human ERBs, using the formula of Glasberg and Moore. """
        return 9.265 * numpy.log(1 + freq_hz / (24.7 * 9.265))

    @staticmethod
    def _erb2freq(n_erb):
        """ Converts human ERBs to Hz, using the formula of Glasberg and Moore. """
        return 24.7 * 9.265 * (numpy.exp(n_erb / 9.265) - 1)



import numpy
import pytest
import slab

def test_signal_generation():
    # (numpy.ndarray | object | list)
    for _ in range(100):
        n_samples = numpy.random.randint(100, 10000)
        n_channels = numpy.random.randint(1, 10)
        samplerate = numpy.random.randint(10, 1000)
        sig = slab.Signal(numpy.random.randn(n_samples, n_channels), samplerate)
        assert sig.n_channels == n_channels
        assert sig.n_samples == n_samples
        assert sig.samplerate == samplerate
        assert len(sig.times) == len(sig.data)
        numpy.testing.assert_almost_equal(sig.times.max()*samplerate, n_samples, decimal=-1)


def test_arithmetics():
    for _ in range(100):
        n_samples = numpy.random.randint(100, 10000)
        n_channels = numpy.random.randint(1, 10)
        samplerate = numpy.random.randint(10, 1000)
        sig = slab.Signal(numpy.random.randn(n_samples, n_channels), samplerate=samplerate)
        numpy.testing.assert_equal((sig+sig).data, (sig*2).data)
        numpy.testing.assert_equal((sig-sig).data, numpy.zeros([n_samples, n_channels]))
        numpy.testing.assert_equal((sig*sig).data, sig**2)


def test_samplerate():
    samples = slab.Signal.in_samples([1.,2], 1000)
    assert samples[0] == 1000
    assert samples[1] == 2
    orig_rate = slab.signal._default_samplerate
    slab.set_default_samplerate(10000)
    assert slab.signal._default_samplerate == 10000
    slab.set_default_samplerate(orig_rate)
    dur_seconds = numpy.abs(numpy.random.randn(100))
    dur_samples = numpy.random.randint(10, 100000, 100)
    for i in range(100):
        samplerate = numpy.random.randint(1, 100000)
        assert numpy.abs(slab.Signal.in_samples(dur_seconds[i], samplerate) - int(dur_seconds[i]*samplerate)) <= 1
        assert slab.Signal.in_samples(dur_samples[i], samplerate) == dur_samples[i]


def test_channels():
    for _ in range(100):
        n_samples = numpy.random.randint(100, 10000)
        n_channels = numpy.random.randint(1, 10)
        samplerate = numpy.random.randint(10, 1000)
        sig = slab.Signal(numpy.random.randn(n_samples, n_channels), samplerate=samplerate)
        for i, ch in enumerate(sig.channels()):
            numpy.testing.assert_equal(sig.channel(i).data, ch.data)


def test_resize():
    dur_seconds = numpy.abs(numpy.random.randn(100))
    dur_samples = numpy.random.randint(10, 100000, 100)
    for i in range(100):
        n_samples = numpy.random.randint(100, 10000)
        n_channels = numpy.random.randint(1, 10)
        samplerate = numpy.random.randint(10, 1000)
        sig = slab.Signal(numpy.random.randn(n_samples, n_channels), samplerate=samplerate)
        resized = sig.resize(dur_seconds[i])
        assert numpy.abs(resized.n_samples - dur_seconds[i]*resized.samplerate) < 1
        resized = sig.resize(dur_samples[i])
        assert resized.n_samples == dur_samples[i]


def test_trim():
    for _ in range(100):
        n_samples = numpy.random.randint(100, 10000)
        n_channels = numpy.random.randint(1, 10)
        samplerate = numpy.random.randint(10, 1000)
        sig = slab.Signal(numpy.random.randn(n_samples, n_channels), samplerate=samplerate)
        start, stop = numpy.sort(numpy.random.randint(0, n_samples+1, 2))
        if start == stop:
            start -= 1
        if numpy.random.rand() < 0.5:
            trimmed = sig.trim(start=start, stop=stop)
        else:
            trimmed = sig.trim(start=start/samplerate, stop=stop/samplerate)
        assert numpy.abs(trimmed.n_samples - (stop-start)) <= 1
    with pytest.raises(ValueError):  # testing start not preceding stop case
        trimmed = sig.trim(start=stop, stop=start)


def test_resample():
    for _ in range(100):
        sig = slab.Signal(numpy.random.randn(numpy.random.randint(100, 10000)))
        samplerate = numpy.random.randint(10, 1000)
        sig_resampled = sig.resample(samplerate)
        assert numpy.abs(sig_resampled.n_samples - sig.duration*samplerate) < 1


def test_envelope():
    sig = slab.Sound.tone()
    _ = sig.envelope(kind="gain")
    _ = sig.envelope(kind="dB")
    _ = sig.envelope(apply_envelope=[0, 1, 0], times=[0, 0.3, 1])
    for _ in range(100):
        env = numpy.random.uniform(-1, 1, 3)
        sig2 = sig.envelope(apply_envelope=env)
        assert numpy.abs(sig2.data.max() - sig.data.max() * numpy.abs(env).max()) < .001


def test_delay():
    dur_seconds = numpy.abs(numpy.random.randn(100))
    dur_samples = numpy.random.randint(10, 100000, 100)
    for i in range(100):
        sig = slab.Signal(numpy.random.randn(numpy.random.randint(100, 10000), numpy.random.randint(1, 10)))
        channel = numpy.random.choice(sig.n_channels)
        for dur in [dur_samples[i], dur_seconds[i]]:
            delay_n_samples = slab.Signal.in_samples(dur, slab.signal._default_samplerate)
        if delay_n_samples > sig.n_samples:
            numpy.testing.assert_raises(ValueError, sig.delay, dur, channel)
        else:
            delayed = sig.delay(duration=dur, channel=channel, filter_length=numpy.random.randint(100, 10000)*2)
            delayed_channel = delayed.channel(channel)
            numpy.testing.assert_almost_equal(delayed_channel[:delay_n_samples].flatten(),
                                              numpy.zeros(delay_n_samples), decimal=7)
